"""
# Example code for creating a data_handler
# populating it by importing a npy dataset
# setting mapping and parents
# and writing the data to Maya

import fmc_trackdata_handler as df
import fmc_maya_tools as mt
import openpose_data_mappings as data_mapping

### Make empty data structure
data_handler = df.FmcTracDataHandler()

### Set mapping
data_handler.set_parent_mapping(data_mapping.parent_mapping)
data_handler.set_point_name_mapping(data_mapping.point_name_mapping)

### Load openpose npy data into variable
import numpy as np
datafile_path = r"G:\Shared drives\000_projects\cl005_northeasternMarkerlessMocap\incoming\datasets\sesh_21-05-31_111833\DataArrays\openPoseSkel_3d.npy"
data = np.load(datafile_path)

### Import data into data handler as an actor. Handler can contain multiple actors
# This step can be repeated for multiple datasets, adding each as an individual actor
actor_name = "tester"
data_handler.import_actor_raw_data(actor_name, data)

### Write content of data_handler to Maya
mt.write_to_maya(data_handler, slice=False)
"""

import maya.cmds as cmds
import math


def write_to_maya(fmcDataObj, slice=False, sample_by = 5):
    """
    Takes a trackdata handler object and creates animated maya objects accordingly

    input:
        fmcDataObj: object og type FmcTracDataHandler
        slice: if not false should be an int tuple of len=2 defining start and end frame to write to Maya
        sample_by: only write every n'th frame to Maya
    """
    for actor_name in fmcDataObj.list_actors():
        ### Creating point_names
        sample_length = fmcDataObj.get_actor_sample_count(actor_name)
        tracking_points = fmcDataObj.get_actor_tracking_points(actor_name)
        point_names = tracking_points.keys()

        # Create point locators
        pnt_locs = {}
        for pnt_name in point_names:
            pnt_locs[pnt_name] = cmds.spaceLocator(n=pnt_name+ "_loc")[0]

        #cmds.scale(0.2, 0.2, 0.2, list(pnt_locs.values()))


        # Make line connector between point locators according to "parents"
        connect_lines = []
        for pnt_name in point_names:
            parents = fmcDataObj.get_point_parents(actor_name, pnt_name)
            for parent_name in parents:
                line = make_line_between(pnt_locs[pnt_name], pnt_locs[parent_name], "connectorLine_" + pnt_name + "_" + parent_name)
                connect_lines.append(line)

        # group everything under the same transform
        grp = cmds.group(list(pnt_locs.values()) + connect_lines, name="mmcap_" + actor_name + "_grp")

        # Adding position data to point locators
        if not slice == False and len(slice)==2:
            print("Slice active: using samples in range: %s:%s" % (slice[0], slice[1]))
            sample_range = range(slice[0], slice[1], sample_by)
        else:
            sample_range = range(0, sample_length, sample_by)

        progress_step = 100./len(sample_range)
        progress = 0
        last_progress = 0
        for sample_i in sample_range:
            for pnt_name in point_names:
                loc_name = pnt_locs[pnt_name]
                pos = tracking_points[pnt_name]["samples"][sample_i]
                if not math.isnan(pos[0]):
                    cmds.setKeyframe(loc_name, v=pos[0], at='translateX', time=sample_i )
                    cmds.setKeyframe(loc_name, v=pos[1], at='translateY', time=sample_i )
                    cmds.setKeyframe(loc_name, v=pos[2], at='translateZ', time=sample_i )
            progress += progress_step
            if int(progress) > int(last_progress):
                print(str(int(progress)).zfill(2) + "%")
                last_progress = progress


        # Reorienting base group to align with grid
        for attr in reorient_transform:
            cmds.setAttr(grp+"." + attr, reorient_transform[attr])

        cmds.playbackOptions(minTime= sample_range[0], maxTime = sample_range[-1])




def make_line_between(obj1, obj2, line_name):
    """
    creates a curve segment and attaches each end to obj1 and obj2 accordingly
    """
    line = cmds.curve(name=line_name, d=1, p=[(0,0,0),(0,0,0)])
    cmds.connectAttr(obj1+".t", line + ".cv[0]")
    cmds.connectAttr(obj2+".t", line + ".cv[1]")
    return line



# Imperically derived orientation to place actor on grid in Maya
reorient_transform = {
    "translateX": 6.976,
    "translateY": -21.502,
    "translateZ": 47.516,
    "rotateX": -114.918,
    "rotateY": 4.391,
    "rotateZ":0,
    "scaleX": 0.031,
    "scaleY": 0.031,
    "scaleZ": 0.031,
    }



from pathlib import Path

from ruamel.yaml import YAML

from freemocap import recordingconfig
from freemocap.fmc_startup import startupGUI

import tkinter as tk
from tkinter import filedialog


def get_user_preferences(session,stage):
    """
    load user preferences if they exist, create a new preferences yaml if they don't
    """
    path_to_this_py_file = Path(__file__).parent
    preferences_path = path_to_this_py_file/'user_preferences.yaml' 
    preferences_yaml = YAML()

    #check for a user preferences yaml, if it doesn't exist, build one using the default parameters in recordingconfig.py   
    if preferences_path.exists():
        preferences = preferences_yaml.load(preferences_path)
    else:
        preferences = recordingconfig.parameters_for_yaml
        preferences_yaml.dump(preferences, preferences_path)
    
    session.preferences = preferences
    session.preferences_path = preferences_path
        
def get_dlc_paths(session):
        try:
            saved_dlc_paths = session.preferences['saved']['dlc_config_paths']
        except: 
            saved_dlc_paths = session.preferences['default']['dlc_config_paths']

        dlc_config_paths = startupGUI.RunChooseDLCPathGUI(session,saved_dlc_paths)
        
        session.preferences['saved']['dlc_config_paths'] = dlc_config_paths
        session.save_user_preferences(session.preferences)
        return dlc_config_paths

def get_data_folder_path(session):
        #if we are rerunning a session folder
        # 1) Check if we're using the last saved dataFolderPath, or if the user wants to choose a different one
        #   a. if the user wants to choose one, bring up a GUI to let them decide
        #   b. if we're using the last known path - parse the user preferences yaml (and check if that yaml exists)
        # 2) Check that the data folder exists
        # 3) If no sessionID was user-input, search the chosen directory for the last session created
        if session.setDataPath == True:
            session.basePath = startupGUI.RunChooseDataPathGUI(session)
            session.preferences['saved']['path_to_save'] = str(session.basePath)
            session.save_user_preferences(session.preferences)
            
            #sesh.dataFolderPath = Path(basePath)/sesh.dataFolderName

        elif session.userDataPath is not None:
            session.basePath = session.userDataPath
        
        else:
            try:
                current_path_to_data = session.preferences['saved']['path_to_save']
                session.basePath = Path(current_path_to_data)
            except KeyError:
                print('Saved Data path not found, please choose a new one')
                session.basePath = startupGUI.RunChooseDataPathGUI(session)
                session.preferences['saved']['path_to_save'] = str(session.basePath)
                session.save_user_preferences(session.preferences)

    
        if session.basePath.stem == session.dataFolderName: #don't recursively craete 'FreeMoCap_Data' folders!
            dataFolder = session.basePath
        else:
            dataFolder = session.basePath/session.dataFolderName
    
        session.dataFolderPath = dataFolder


        if not dataFolder.exists():
            print('Data folder not found at this path. Please choose another path.')
            session.basePath = startupGUI.RunChooseDataPathGUI(session)
            session.preferences['saved']['path_to_save'] = str(session.basePath)
            session.save_user_preferences(session.preferences)

            if session.basePath.stem == session.dataFolderName: #don't recursively craete 'FreeMoCap_Data' folders!
                dataFolder = session.basePath
            else:
                dataFolder = session.basePath/session.dataFolderName
            session.dataFolderPath = dataFolder



       

def get_blender_path(session, resetBlenderExe):
    if  resetBlenderExe == True:
        print('resetting blender.exe path')
        session.preferences['saved']['blenderEXEpath'] = None
    if session.preferences['saved']['blenderEXEpath'] == None:
        print('Please select your blender.exe file')
        root = tk.Tk()
        root.withdraw()
        blender_file_path = filedialog.askopenfilename(title = 'Select your Blender executable (usually somehwere like `C:/Program Files/Blender Foundation/2.93/blender.exe')
        print('Using blender executable located at:', blender_file_path)
        session.preferences['saved']['blenderEXEpath'] = blender_file_path
        session.save_user_preferences(session.preferences)
    else:
        blender_file_path = session.preferences['saved']['blenderEXEpath']
        print('Using blender executable located at:', blender_file_path)

    return blender_file_path




from fastapi import APIRouter
from pydantic import BaseModel

healthcheck_router = APIRouter()


class HealthCheckResponse(BaseModel):
    message: str = "OK"


@healthcheck_router.get('/health')
def route():
    try:
        return HealthCheckResponse()
    except:
        raise ValueError("Unhealthy")




import uvicorn
from fastapi import FastAPI

from src.api.routes import enabled_routers


def create_app():
    app = FastAPI()

    for router in enabled_routers:
        app.include_router(router)

    return app


if __name__ == "__main__":
    app = create_app()
    uvicorn.run(app, host="127.0.0.1", port=5000, log_level="info")



from pathlib import Path
import datetime
import sys
import pandas as pd
import numpy as np

from freemocap.webcam import recordGUI, camsetup
from freemocap import recordingconfig, fmc_anipose
from rich.console import Console

console = Console()

def initialize(session, stage, board):
    """ 
    Runs the initialization needed to start a new recording from scratch (Stage 1)
    Create a new sessionID and attempt to load previous recording parameters (camera settings/rotations/save locations), 
    and use default values if previous parameters are not found. Ask user to select preferences using a series of GUIs, runs the Setup option if 
    chosen by the user, and ultimately sets a bool that dictates whether we should proceed to start the actual Stage 1 recording. Also saves parameters
    into the session class, which gets saved into the user_preferences yaml

    """ 
    
    console.rule(style="color({})".format(13))    
    console.rule('Finding available webcams',style="color({})".format(13))
    console.rule(style="color({})".format(13)) 

    # if stage == 1:
    filepath = Path.cwd()

    session.board = board

    # %% Stage One Initialization


    # create session ID
    sessionID_in = datetime.datetime.now().strftime("sesh_%Y-%m-%d_%H_%M_%S")
    session.sessionID = sessionID_in

#load the parameters YAML and extract the saved parameters if possible, and the default otherwise       
    #here = Path(__file__).parent
    # parameter_path = here/'user_preferences.yaml'
    # parameters_yaml = YAML()
    # if parameter_path.exists():
    #     parameters = parameters_yaml.load(parameter_path)
    # else:
    #     parameters = recordingconfig.parameters_for_yaml
    #     with open(parameter_path,'w') as outfile:
    #         parameters_yaml.dump(parameters, outfile)




    proceedToRecording = False #create this boolean, set it to false, and if the user wants to record
                                #later in the pipeline, it will be set to true

#run the GUI to get the tasks, the cams chosen, the camera settings, and the session ID
    cam_inputs, task = recordGUI.RunChoiceGUI()
    restartSetup = True

    while restartSetup == True:
        
        try:
            rotation_entry = session.preferences["saved"]["rotations"]
            parameter_entry = session.preferences["saved"]["parameters"]
        except:
            print("Could not load saved parameters, using default parameters")
            rotation_entry = session.preferences['default']['rotations']
            parameter_entry = session.preferences['default']['parameters']
                

        try:
            current_path_to_save = session.preferences['saved']['path_to_save']
        except:
            current_path_to_save = session.preferences['default']['path_to_save'] 
        rotDict, paramDict, session.sessionID,savepath, mediaPipeOverlay = recordGUI.RunParametersGUI(sessionID_in, rotation_entry, parameter_entry, current_path_to_save, cam_inputs, task)
    
    #update the saved parameters in the YAML
        #recordingconfig.rotation_settings['saved'] = rotDict
        #recordingconfig.camera_session.preferences['saved'] = paramDict
        session.preferences['saved']['rotations'] = rotDict
        session.preferences['saved']['parameters'] = paramDict
        if savepath is not None:
            session.preferences['saved']['path_to_save'] = savepath

        session.save_user_preferences(session.preferences)

    #save recording parameters to the config yaml

        
    #create a list from the rotation dictionary to be used in running webcams
        rotation_input = list(rotDict.values())

        if task == "setup":
            # run setup processes, and then check if th user wants to proceed to recording
            camsetup.RunSetup(cam_inputs, rotation_input, paramDict,mediaPipeOverlay)
            proceedToRecording, restartSetup, session.sessionID, savepath = recordGUI.RunProceedtoRecordGUI(
                sessionID_in,current_path_to_save
            )
            session.preferences['saved']['path_to_save'] = savepath
            session.save_user_preferences(session.preferences)
        elif task == "record":
            proceedToRecording = True
            restartSetup = False

    if proceedToRecording:
        # create these session properties to be used later in the pipeline
        session.cam_inputs = cam_inputs
        session.parameterDictionary = paramDict
        session.rotationInputs = rotation_input
        session.basePath = Path(savepath)

        #create a config yaml and text file for this session
        session.start_session(session.parameterDictionary,session.rotationInputs)
        session.session_settings['recording_parameters']['RotationInputs'] = rotDict
        session.session_settings['recording_parameters']['ParameterDict'] = paramDict
        #configyaml = YAML()
        #sessionsettings = configyaml.load(session.session_yaml_path)
        #sessionsettings['recording_parameters']['RotationInputs'] = rotDict
        #sessionsettings['recording_parameters']['ParameterDict'] = paramDict
        #configyaml.dump(sessionsettings, session.session_yaml_path)
        #print(sessionsettings)
        #config_yaml_path = recordingconfig.createSession(session,filepath)
        #recordingconfig.createSessionTxt(session,paramDict,rotDict)
        
        print('Proceeding to Stage One')        

    else:
    
        sys.exit('Recording Canceled')



##
##
##  ███████ ███    ███  ██████          ██████  █████  ███    ███ ███████ ██████   █████  
##  ██      ████  ████ ██              ██      ██   ██ ████  ████ ██      ██   ██ ██   ██ 
##  █████   ██ ████ ██ ██              ██      ███████ ██ ████ ██ █████   ██████  ███████ 
##  ██      ██  ██  ██ ██              ██      ██   ██ ██  ██  ██ ██      ██   ██ ██   ██ 
##  ██      ██      ██  ██████          ██████ ██   ██ ██      ██ ███████ ██   ██ ██   ██ 
##
##
#Font - ANSI Regular - https://patorjk.com/software/taag/#p=display&f=ANSI%20Regular&t=Play%20Skeleton%20Animation
                                                                                      
                                                                                      

import platform
from pathlib import Path
import time

import cv2
import numpy as np

#python rich stuff - https://rich.readthedocs.io/
from rich import print
from rich.console import Console



cap_default_parameters_dict = {
                            'exposure': -5,
                            'cap_resolution_width': 1280,
                            'cap_resolution_height': 720,
                            }

class FMC_Camera:
    """ Class to open a cv2.VideoCapture object, with parts to connect to a `FMC_MutliCam` object as part of a FMC_Session. 
    Note, this class was made in service of the needs of the FreeMoCap project, at some point I'd love to develop this into a fully generic and useful 'camera' object, right now it's designed with that application (and only that application) in mind
    """
    ##  
    ##  
    ##                  ██ ███    ██ ██ ████████                 
    ##                  ██ ████   ██ ██    ██                    
    ##                  ██ ██ ██  ██ ██    ██                    
    ##                  ██ ██  ██ ██ ██    ██                    
    ##  ███████ ███████ ██ ██   ████ ██    ██    ███████ ███████ 
    ##                                                           
    ##  
    ##  

    def __init__(
                self, 
                cam_num=0, 
                rotation_code=None,
                frame_queue = None,
                barrier = None,
                exit_event = None,
                show_cam_stream = False,  
                rich_console = None,       
                show_console = False,       
                cap_parameters_dict = cap_default_parameters_dict,
                vid_save_path=None,  
                ):
        """Open a camera stream using `cv2.VideoCapture()`. 
        If `Windows`, use `cv2.VideoCapture(camNum, cv2.CAP_DSHOW)` to make video initialize much faster. Otherwise, use `cv2.VideoCapture(camNum, cv2.CAP_ANY)

        Args:
            camNum (int, optional): The ID of the camera to open. Defaults to 0 (i.e. the first camera it finds)
            vid_save_path (pathlib Path, optional): Where to save the video. Defaults to None.
            barrier (threading.Barrier, optional): Barrier for syncronized multicam recording. Defaults to None.
            frame_queue (queue.Queue, optional): [description]. Queue to send out images when we're recording in a thread.
        """

        self._cam_num = cam_num
        self._cam_name = 'Camera_{}'.format(str(self._cam_num).zfill(2))
        self._cam_short_name = 'cam{}'.format(self._cam_name)
        
        self._vid_cap_start_time_unix = None

        self._cap_parameters_dict = cap_parameters_dict
        
        self._rotation_code = rotation_code

        self._frame_queue = frame_queue
        self._barrier = barrier
        self._exit_event = exit_event
        self._vid_save_path = vid_save_path
        self._show_console = show_console
        self.rich_console = rich_console # a console object from the Rich python package


        if vid_save_path:
            self._vid_save_path = Path(vid_save_path)
        

        if not self.rich_console:
            self.rich_console = Console()
        
        self.open() #open VideoCapture Object

        self._show_cam_stream = show_cam_stream
        if self._show_cam_stream:
            self.show()#display camera stream

        if self._frame_queue: 
            self.run_in_thread() #launch in thread (or process?) mode
    ##   
    ##   ██████  ██████   ██████  ██████  ███████ ██████  ████████ ██ ███████ ███████ 
    ##   ██   ██ ██   ██ ██    ██ ██   ██ ██      ██   ██    ██    ██ ██      ██      
    ##   ██████  ██████  ██    ██ ██████  █████   ██████     ██    ██ █████   ███████ 
    ##   ██      ██   ██ ██    ██ ██      ██      ██   ██    ██    ██ ██           ██ 
    ##   ██      ██   ██  ██████  ██      ███████ ██   ██    ██    ██ ███████ ███████ 
    ##                                                                                
    ##                                                                                
    @property
    def cam_num(self):
        return self._cam_num
    
    @property
    def cam_name(self):
        return self._cam_name
    
    @property
    def cam_short_name(self):
        return self._cam_short_name

    @property
    def vid_save_path(self):
        return self._vid_save_path 

    @property
    def show_cam_stream(self):
        return self._show_cam_stream
    
    @show_cam_stream.setter
    def show_cam_stream(self, input_bool):
        self._show_cam_stream = input_bool

    @property
    def vid_cap_start_time_unix(self):
        return self._vid_cap_start_time_unix

    @property
    def vid_cap_timestamps_unix_ns(self): #not really a property though?
        return self._vid_cap_timestamps_unix_ns

    @property
    def show_console(self):
        return self._show_console
    
    @show_console.setter
    def show_console(self, input_bool):
        self._show_console = input_bool
    
    ###
    ###   
    ###   ███    ███ ███████ ████████ ██   ██  ██████  ██████  ███████     ███    ███  █████  ███    ██ 
    ###   ████  ████ ██         ██    ██   ██ ██    ██ ██   ██ ██          ████  ████ ██   ██ ████   ██ 
    ###   ██ ████ ██ █████      ██    ███████ ██    ██ ██   ██ ███████     ██ ████ ██ ███████ ██ ██  ██ 
    ###   ██  ██  ██ ██         ██    ██   ██ ██    ██ ██   ██      ██     ██  ██  ██ ██   ██ ██  ██ ██ 
    ###   ██      ██ ███████    ██    ██   ██  ██████  ██████  ███████     ██      ██ ██   ██ ██   ████ 
    ###                                                                                                 
    ###
    ###                                                                                                 

    def show_cam_stream(self, show_cam_stream_bool=True):
        """whether to open a window showing camera stream

        calls `self.show()` when set to `True`
        and  calls `self.close()` when set to `False` (b/c I'm too lazy to figure out how to make ti close th the display without closing the capture right now, sorry lol)

        Args:
            show_cam_stream_bool (Bool): whether to show stream (default)

        Returns:
            self
        """
        assert type(show_cam_stream_bool) == type(True), 'Input to `show_cam_stream` must be a bool'
        self._show_cam_stream = show_cam_stream_bool

        if self._show_cam_stream:
            self.show()
        else:
            self.close()

        return self._show_cam_stream
    ###   
    ###    ██████  ██████  ███████ ███    ██     ██    ██ ██ ██████       ██████  █████  ██████  
    ###   ██    ██ ██   ██ ██      ████   ██     ██    ██ ██ ██   ██     ██      ██   ██ ██   ██ 
    ###   ██    ██ ██████  █████   ██ ██  ██     ██    ██ ██ ██   ██     ██      ███████ ██████  
    ###   ██    ██ ██      ██      ██  ██ ██      ██  ██  ██ ██   ██     ██      ██   ██ ██      
    ###    ██████  ██      ███████ ██   ████       ████   ██ ██████       ██████ ██   ██ ██      
    ###                                                                                          
    ###                                                                                          
    def open(self):
        """
        Open Video Capture Object and dispaly in little video
        """

        if platform.system() == 'Windows':
            self.cv2_cap = cv2.VideoCapture(self._cam_num, cv2.CAP_DSHOW)
        else:
            self.cv2_cap = cv2.VideoCapture(self._cam_num, cv2.CAP_ANY)

        self.cv2_cap.set(cv2.CAP_PROP_EXPOSURE, self._cap_parameters_dict['exposure'])
        self.cv2_cap.set(cv2.CAP_PROP_FRAME_WIDTH, self._cap_parameters_dict['cap_resolution_width'])
        self.cv2_cap.set(cv2.CAP_PROP_FRAME_HEIGHT, self._cap_parameters_dict['cap_resolution_height'])

        self.cv2_cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')) 


        self.video_exposure = self.cv2_cap.get(cv2.CAP_PROP_EXPOSURE)
        self.video_resolution_width = self.cv2_cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        self.video_resolution_height = self.cv2_cap.get(cv2.CAP_PROP_FRAME_HEIGHT)

        if not self.cv2_cap.isOpened():
            self.rich_console.log("Camera# "+str(self._cam_num)+" failed to open :(")
            
        else:
            self.rich_console.log("Camera# "+str(self._cam_num)+"  started - Exposure:({}) - Resolution(width, height):({},{}) ".format(self.video_exposure, self.video_resolution_width, self.video_resolution_height))
            self._vid_cap_start_time_unix = time.time_ns() #the precision is aspirational, lol
            self._vid_cap_timestamps_unix_ns = np.empty(0)
    
    ###
    ###  ███████ ██   ██  ██████  ██     ██     ██    ██ ██ ██████  
    ###  ██      ██   ██ ██    ██ ██     ██     ██    ██ ██ ██   ██ 
    ###  ███████ ███████ ██    ██ ██  █  ██     ██    ██ ██ ██   ██ 
    ###       ██ ██   ██ ██    ██ ██ ███ ██      ██  ██  ██ ██   ██ 
    ###  ███████ ██   ██  ██████   ███ ███        ████   ██ ██████  
    ###
                                                                                                                                                          
    def show(self):
        """
        display the camera stream, press ESC to close (and release the `cv2_cap`)
        """
        success, image, timestamp = self.read_next_frame()
                    
        while success:
            cv2.imshow(self._cam_name, image)
            
            success, image, timestamp = self.read_next_frame()

            console_msg = self.cam_name +  " read an image at {:.4f}".format(timestamp)
            self.update_console(console_msg)
            self._vid_cap_timestamps_unix_ns = np.append(self.vid_cap_timestamps_unix_ns, timestamp)

            if self.wait_key() == 27:  # exit on ESC                        
                self.close()
            
            if cv2.getWindowProperty(self._cam_name, cv2.WND_PROP_VISIBLE) < 1: #break loop if window closed
                break   
        cv2.destroyAllWindows()

    ###  
    ###  ██████  ██    ██ ███    ██     ██ ███    ██     ████████ ██   ██ ██████  ███████  █████  ██████  
    ###  ██   ██ ██    ██ ████   ██     ██ ████   ██        ██    ██   ██ ██   ██ ██      ██   ██ ██   ██ 
    ###  ██████  ██    ██ ██ ██  ██     ██ ██ ██  ██        ██    ███████ ██████  █████   ███████ ██   ██ 
    ###  ██   ██ ██    ██ ██  ██ ██     ██ ██  ██ ██        ██    ██   ██ ██   ██ ██      ██   ██ ██   ██ 
    ###  ██   ██  ██████  ██   ████     ██ ██   ████        ██    ██   ██ ██   ██ ███████ ██   ██ ██████  
    ###                                                                                                   
    ###                                                                                                   
    def run_in_thread(self):
        """run camera in a thread
         put incoming images into an tuple containing camNum_image_timestamp_tuple and stuff it into `self._frame_queue`
        """
        # with self.rich_console.status('Camera# {} is running'.format(self.cam_num)):
        while not self._exit_event.is_set():
            try:
                success, image, timestamp = self.read_next_frame()
                cam_image_timestamp_tuple = (self.cam_num, image, timestamp)
                            
                self._frame_queue.put(cam_image_timestamp_tuple) #stuff this frame tuple packet into this camera's Queueue object

                # log_msg = self.cam_name+" got a frame at timestamp:"+ str(timestamp) + ' queue size: ' + str(self._frame_queue.qsize())
                # self.rich_console.log(log_msg)

                self._barrier.wait() #wait until other cams have grabbed a frame (and the frame_grabber has frame grabbed them)
            except:
                self.rich_console.print_exception()
        
        #shut down when 'exit_event' is tripped
        self.close()
        

    def run_in_subprocess(self):
        self.run_in_thread()

    ###     
    ###         
    ###     ███████ ████████  ██████          
    ###     ██         ██    ██               
    ###     █████      ██    ██               
    ###     ██         ██    ██               
    ###     ███████    ██     ██████ ██ ██ ██ 
    ###     
    ###     
    

    
    def __enter__(self):
        """Context manager -  No need to do anything special on start"""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Context manager -  close `self.cv2_cap` on exit"""         
        self.close()
        return self

    def update_console(self, console_msg):
        if self.show_console:
            self.rich_console.log(console_msg) #there should be a way to log this without printing it, but I don't know how....

    def read_next_frame(self):
        """ Grab next frame, return `Tuple(success[bool, True], image, timestamp)` on success or `Tuple(success[bool, False], False, False)` on failure

        Returns:
            Success[bool]: True if frame image captured successfully
            image: the image that was recieved from the VideoCapture object
            timestamp: unix timestamp from time.time_ns()
        """
        success, image = self.cv2_cap.read()
        timestamp = time.time_ns() #the precision is aspirational, lol
        if success:
            self._vid_cap_timestamps_unix_ns = np.append(self._vid_cap_timestamps_unix_ns, time.time_ns()) #the precision is aspirational, lol

            if self._rotation_code is None:
                pass
            elif self._rotation_code == 'cv2.ROTATE_90_CLOCKWISE':                
                image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
            elif self._rotation_code == 'cv2.ROTATE_90_COUNTERCLOCKWISE':                
                image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
            elif self._rotation_code == 'cv2.ROTATE_180':                
                image = cv2.rotate(image, cv2.ROTATE_180)
            else:
                Exception, 'Invalid rotation code entered for FMC_Camera: {}'.format(self.cam_name)

            return success, image, timestamp
        else:
            return success, success, success
    
    def close(self):
        """
        close the video
        """
        self.cv2_cap.release()
        self.rich_console.log("Camera# "+str(self._cam_num)+" shutting down")

    
    def wait_key(self):

        if self.show_cam_stream:
            self._wait_key = cv2.waitKey(1)            
        else: 
            self._wait_key = 27

        return self._wait_key


### 
### ██ ███████                     ███    ███  █████  ██ ███    ██                 
### ██ ██                          ████  ████ ██   ██ ██ ████   ██                 
### ██ █████                       ██ ████ ██ ███████ ██ ██ ██  ██                 
### ██ ██                          ██  ██  ██ ██   ██ ██ ██  ██ ██                 
### ██ ██          ███████ ███████ ██      ██ ██   ██ ██ ██   ████ ███████ ███████ 
### 
###                                                                                                                                                               
if __name__ == '__main__':

    with FMC_Camera() as cam:
        console = Console() #create Rich console to catch and print exceptions
        try:
            while True:
                cam.show()
                break
        except Exception:
            console.print_exception()




import bpy 
import pickle

from pathlib import Path

class FMC_OT_loadVideos(bpy.types.Operator):
    """load in each of the nSynced Vidoes (n= numCams) as planes and place in 3d scene"""
    bl_idname = 'fmc.load_nsynced_videos'
    bl_label = "Load nSynched Videos"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        session_path = Path(context.scene.fmc_session_path)
                         
        if not session_path.is_dir():
            session_path = session_path.parent
        
         
        camera_calibration_pickle_path = session_path / (session_path.name + '_calibration.pickle')
        
        vidFolderPath = session_path / 'SyncedVideos'
        
           

        with open(camera_calibration_pickle_path, 'rb') as camera_calib_pickle:
            camera_calibration_dict = pickle.load(camera_calib_pickle)
  
        camera_translations = []
        camera_rotations = []
        for this_cam_calib_info in camera_calibration_dict:
            camera_translations.append(this_cam_calib_info['translation'])
            camera_rotations.append(this_cam_calib_info['rotation'])
        
 
        
        for vid_number, thisVidPath, in enumerate(vidFolderPath.glob('*.mp4')): 
            
            this_camera_tx = camera_translations[vid_number][0]/1000 #convert to meters from millimeters
            this_camera_ty = camera_translations[vid_number][1]/1000
            this_camera_tz = camera_translations[vid_number][2]/1000
            
            this_camera_rx = camera_rotations[vid_number][0]
            this_camera_ry = camera_rotations[vid_number][1]
            this_camera_rz = camera_rotations[vid_number][2]
            
            #create camera object on Z-axis pointing down at origin
            bpy.ops.object.camera_add(align='WORLD')
            thisCam = context.active_object
            thisCam.location = (this_camera_tx,this_camera_ty,this_camera_tz)
            thisCam.rotation_euler = (this_camera_rx, this_camera_ry, this_camera_rz)
            # thisCam.scale = (5, 5, 5)
            thisCam.name = 'Cam' + str(iter)
            
            # use 'images as planes' add on to load in the video files as planes 
            bpy.ops.import_image.to_plane(files=[{"name":thisVidPath.name}], directory=str(thisVidPath.parent), )
            thisVidPlane = context.active_object
            thisVidPlane.parent = thisCam #parent this image_plane to camera object
            thisVidPlane.location[2] -= 2 #bump it down a bit for aesthetics

            # context.active_object.location = (iter,iter,iter) #bump this image_plane over a bit so they don't over lap
            
        return {'FINISHED'}


import numpy as np
import pandas as pd
import sys
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import tkinter as tk
from tkinter import Tk, Label, Button, Frame

def TimeSync(session,timeStampData,numCamRange,camNames):    
    """
    First creates a master timeline from the average frame length interval across all cameras. Then proceeds to compare timepoints in master timeline
    to timestamps in each camera and find the closest matching timestamp in a camera to a point on the master timeline and create a list of
    frame numbers.

    Then loop through the list of frame numbers from each camera. Update counts for missing frames and duplicated frames. Create a table with these counts
    and plots of statistics related to the recording
    """     
    def CloseNeighb(camera,point): 
        """ 
        Find the timestamp in the camera timestamp list that most closely matches the input pointin the master timeline
        """  
        closestPoint = (np.abs(camera - point)).argmin() 
        return closestPoint

    # this section auto-finds the start and end points for our master timeline to the nearest second
    masterTimelineBegin = np.ceil(max(timeStampData.iloc[0]))
    lastPoints = []
    for name in camNames:
        cameraLastPoint = timeStampData[name][
            timeStampData[name].last_valid_index()
        ]  # find the last non-nan value in each camera
        lastPoints.append(cameraLastPoint)
    
    masterTimelineEnd = np.floor(min(lastPoints))  #find where the fastest camera ended, round down, use that as the end point


    #print('end',endList)
    print('intervals:',masterTimelineBegin,masterTimelineEnd)   
    
    preSyncTotalNumFrames = []
    totalFrameIntvl= [] #list for storing frame rates for each camera
    
      
    n = 0; #counter for going through camera names
    for x in numCamRange:
      currentCam = timeStampData[camNames[x]]
      #finds the closest point in each camera to the start and end points of our timeline
      currentCamStart = CloseNeighb(currentCam,masterTimelineBegin) 
      currentCamEnd = CloseNeighb(currentCam,masterTimelineEnd)
    

      print("using frames:", currentCamStart,"-",currentCamEnd,"for",camNames[n])
      n +=1
      currentCamTimeline = currentCam[currentCamStart:currentCamEnd] #grab the times from start to finish for each camera

      currentCamNumFrames = len(currentCamTimeline)

      currentCamFrameInterval = np.mean(np.diff(currentCamTimeline)) #calculate the interval between each frame and take the mean 
      
      totalFrameIntvl.append(currentCamFrameInterval) #add interval to list 
      preSyncTotalNumFrames.append(currentCamNumFrames)
      #print(camNames[x],currentCamStart,currentCamEnd)

    #print(preSyncTotalNumFrames, 'presync')
    #assert preSyncTotalNumFrames.count(preSyncTotalNumFrames[0]) == len(preSyncTotalNumFrames), "Number of frames sliced for each camera pre-syncing is not the same"

    totalAverageIntvl = np.mean(totalFrameIntvl) #find the total average interval across all cameras
    masterTimeline = np.arange(masterTimelineBegin,masterTimelineEnd,totalAverageIntvl) #build a master timeline with the average interval
       
    #now we start the syncing process
    
    frameList = masterTimeline #start a list of frames, with the first row being our master timeline
    timeList = masterTimeline #start a list of timestamps,with the first row being our master timeline
     
    delNum= []; #stores number of deleted frames/camera
    bufNum= []; #stores number of buffered frames/camera
    bufPercentList = []; #stores percentage of deleted frames/camera
    delPercentList = []; #stores percentage of buffered frames/camera
    
    count = 0 #Keeps track of what frame we're on (I think?)
    n = 0; #I don't remember why I did this but I'm sure I'll figure it out later
    

    postSyncTotalNumFrames = []
    for y in numCamRange:
        thisCam = timeStampData[camNames[y]]
        camTimes = []
        # stored times
        camFrames = []
        # stored frames

        count += 1
        # --------------Adjust each camera to start at the first point in the master timeline
        beginFrame = CloseNeighb(thisCam, masterTimeline[0])
        timeDif = masterTimeline[0] - thisCam[beginFrame]
        thisCam = thisCam + timeDif
        
        for z in masterTimeline: #for each point in the master timeline
            closestFrame = CloseNeighb(thisCam, z) #find the closest frame in this camera to each point of the master timeline
            camFrames.append(closestFrame) #add that frame to the list
            camTimes.append(thisCam[closestFrame]) #find the time corresponding to this frame
            
        print("starting detection:",camNames[n]) #now to start finding deleted/buffered slides
        frameList = np.column_stack((frameList,camFrames)) #update our framelist with our new frames
        timeList = np.column_stack((timeList,camTimes)) #update our timelist 
        
    
        #start counters for the number of buffered and deleted slides
        bufCount = 0
        delCount = 0

        thisCamNumFrames = len(camFrames)
        postSyncTotalNumFrames.append(thisCamNumFrames)

        for i in range(0,len(camFrames)-1): 
            distanceBetweenFrames = camFrames[i+1] - camFrames[i] #find the distance between adjacent frames
            if distanceBetweenFrames == 1: #these frames are consecutive, do nothing
                None 
            elif distanceBetweenFrames == 0: #we have a buffered slide
                bufCount += 1
                # this section looks at our two timepoints, and finds the which point on the master timeline is closest
                frame1 = abs(masterTimeline[i] - camTimes[i])
                frame2 = abs(masterTimeline[i + 1] - camTimes[i])
                # finds which frame is closest, and sets the other frame as a buffer (indicated by the -1)
                if frame1 > frame2:
                    frameList[i, count] = -1
                    timeList[i, count] = -1
                else:
                    frameList[i + 1, count] = -1
                    timeList[i + 1, count] = -1
            elif distanceBetweenFrames > 1:  # deletion
                delCount += 1
            else:

                print("something else happened")
        #update and calculate our percentages and numbers for buffers/deletions
        delNum.append(round(delCount,1))
        bufNum.append(round(bufCount,1))
        #print(bufCount,len(frameList))
        bufPercent = (bufCount/len(frameList))*100
        delPercent = (delCount/len(frameList))*100
        bufPercentList.append(round(bufPercent,2))
        delPercentList.append(round(delPercent,2))
       
        #print(delNum,bufNum,bufPercentList,delPercentList)
        #print("deleted frames:",delCount)
        #print("buffered frames:",bufCount)
        n +=1

    #print(postSyncTotalNumFrames, 'post-sync')
    assert postSyncTotalNumFrames.count(postSyncTotalNumFrames[0]) == len(postSyncTotalNumFrames), "Number of frames sliced for each camera post-syncing is not the same"

    session.postSyncNumFrames = postSyncTotalNumFrames[0]

    #create our data frame for both times and frames    
    frameTable = pd.DataFrame(frameList)   
    columnNames = ['Master Timeline'] + camNames
    frameTable.columns = columnNames
    timeTable = pd.DataFrame(timeList)
    timeTable.columns = columnNames
    totalFrameRate = [round(1 / intvl, 1) for intvl in totalFrameIntvl]
    frameRate = 1 / totalAverageIntvl  # calculates our framerate
    results = {
        "Cam": camNames,
        "#Del": delNum,
        "%Del": delPercentList,
        "#Buf": bufNum,
        "%Buf": bufPercentList,
        "FPS": totalFrameRate,
    }
    resultTable = pd.DataFrame(
        results, columns=["Cam", "#Del", "%Del", "#Buf", "%Buf", "FPS"]
    )

    # ============================================== plot data
    differenceFrame = timeStampData.diff(axis=0)
    fpsFrame = differenceFrame.pow(-1)
    fig = Figure(figsize=[10, 10])
    fig.patch.set_facecolor("#F0F0F0")
    a = fig.add_subplot(221)
    a.set_xlabel("Frame")
    a.set_ylabel("Time(s)")
    timeStampData.plot(ax=a, title="Camera timestamps")
    b = fig.add_subplot(222)
    b.set_xlabel("Frame")
    b.set_ylabel("Frame duration (s)")
    differenceFrame.plot(
        ax=b, marker=".", linestyle="none", title="Camera frame duration"
    )
    c = fig.add_subplot(223)
    c.set_xlabel("Time (s)")
    differenceFrame.plot.hist(
        ax=c, bins=6, alpha=0.5, xlim=[0, 0.1], title="Frame interval distribution (s)"
    )
    d = fig.add_subplot(224)
    d.set_xlabel("Frames per second")
    fpsFrame.plot.hist(
        ax=d, bins=6, alpha=0.5, xlim=[0, 40], title="Frames per second distribution"
    )
    return frameTable, timeTable, frameRate, resultTable, fig


# function to trim our videos


class proceedGUI:
    """ 
    GUI that displays all statistics from recording, and then asks for user input on whether to proceed with the rest of the pipeline
    """  
    def __init__(self, master, results, figure):
        self.master = master
        self.results = results
        self.figure = figure
        self.proceed = ""
        master.title("Choose File Path")

        bottom_frame = Frame(self.master, height=1000)
        bottom_frame.pack(side=tk.LEFT)
        self.label = Label(bottom_frame, text=results.to_string(index=False))
        self.label.pack(side=tk.TOP)
        self.go_button = Button(bottom_frame, text="Proceed", command=self.destroy)
        self.stop_button = Button(bottom_frame, text="Quit", command=self.stop)

        self.canvas = FigureCanvasTkAgg(self.figure, master=self.master)
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(side=tk.LEFT, fill="both", expand=1)

        self.stop_button.pack(side=tk.RIGHT)
        self.go_button.pack(side=tk.RIGHT)

    def stop(self):
        self.master.destroy()
        sys.exit("Quitting Program")

    def destroy(self):
        self.proceed = True
        self.master.destroy()




def time_sync_initialize(session):
    """
    Initialize settings specific to the time syncing stage if running stage 2
    """
    #from the config settings add the camera input parameters to parameter dictionary
    session.initialize(stage = 2)
    session.parameterDictionary = session.session_settings['recording_parameters']['ParameterDict']
    rotationDict = session.session_settings['recording_parameters']['RotationInputs']
    session.rotationInputs = list(rotationDict.values())
    #initialize the path to the timestamp csv
    csvName = session.sessionID + '_timestamps.csv'
    csvPath = session.rawVidPath/csvName

    #read CSV data, turn it into a data frame
    timeStampData = pd.read_csv (csvPath)
    timeStampData = timeStampData.iloc[:,1:]

    #get the camIDs and number of cameras (numCamRange) from the dataframe
    camIDs = list(timeStampData.columns)
    numCams = len(camIDs)
    numCamRange = range(len(camIDs)) 
    
    #create names for each of the raw videos  
    vidNames = []
    for x in numCamRange:
        singleVidName = 'raw_cam{}.mp4'.format(x+1)
        vidNames.append(singleVidName)    


    #initialize all the session variables we'll need to run the rest of the pipeline
    session.timeStampData = timeStampData
    session.camIDs = camIDs
    session.numCamRange = numCamRange
    session.vidNames = vidNames
    session.numCams = numCams



from pathlib import Path


import numpy as np
from rich.progress import Progress
import cv2
import mediapipe as mp

# from numba import jit

mp_drawing = mp.solutions.drawing_utils
mp_drawing_styles = mp.solutions.drawing_styles
mp_holistic = mp.solutions.holistic


def runMediaPipe(session, dummyRun=False):
    """ 
    Run MediaPipe on synced videos, and save body tracking data to be parsed 
    """  



    with mp_holistic.Holistic(
                            static_image_mode = False, #use 'static image mode' to avoid system getting 'stuck' on ghost skeletons?
                            model_complexity = 2, #in this house, we turn the Speed/Accuracy dial all the way towards accuracy \o/)
                            ) as holistic:

        eachCamerasData = []  # Create an empty list that holds each cameras data
        eachCameraResolution = {'Height':[],'Width':[]}
        for (  thisVidPath ) in ( session.syncedVidPath.iterdir() ):  # Run MediaPipe 'Holistic' (body, hands, face) tracker on each video in the raw video folder
            if ( thisVidPath.suffix.lower() == ".mp4" ):  # NOTE - at some point we should build some list of 'synced video names' and check against that

                
                mediaPipe_dataList = []  # Create an empty list for mediapipes data
                if not dummyRun:
                    cap = cv2.VideoCapture(str(thisVidPath))

                    frameNum = -1
                    numFrames = cap.get(cv2.CAP_PROP_FRAME_COUNT)
                    success, image = cap.read()  # load first image from video

                    with Progress() as progress:
                        progressBar = progress.add_task(
                            "[magenta]MediaPiping: {}".format(thisVidPath.name),
                            total=numFrames,
                        )  # make progressbar

                        while success:

                            if frameNum % 5 == 0:
                                progress.update(
                                    progressBar, advance=5
                                )  # increment progress bar everyh 5th frame

                            frameNum += 1

                            image_height, image_width, _ = image.shape

                            results = holistic.process(
                                cv2.cvtColor(image, cv2.COLOR_BGR2RGB), #Convert the BGR image to RGB before processing
                            )  # NOTE: THIS IS WHERE THE MAGIC HAPENS 

                            mediaPipe_dataList.append(
                                results
                            )  # Append data to mediapipe data list

                            # load the next image (will break `while-loop` on the last frame)
                            success, image = cap.read()  # load next image from video

                eachCameraResolution["Height"].append(image_height)
                eachCameraResolution["Width"].append(image_width)

                eachCamerasData.append(
                    mediaPipe_dataList
                )  # Append that cameras data for every frame to the camera datalist

    session.image_height = image_height
    session.image_width = image_width
    session.mediaPipeData = eachCamerasData
    session.eachCameraResolution = eachCameraResolution



def parseMediaPipe(session):
    """ 
    Parse through saved MediaPipe data, and save out a numpy array of 2D points
    """  
    numCams = len(session.mediaPipeData)  # Get number of cameras
    numFrames = len(session.mediaPipeData[0])  # Get number of frames
    # numBodyPoints = len(np.max(session.mediaPipeData[0][:].pose_landmarks.landmark[:]))#Get number of body points
    # numFacePoints = len(np.max(session.mediaPipeData[0][:].face_landmarks.landmark[:]))#Get number of face points
    # numLeftHandPoints = len(np.max(session.mediaPipeData[0][:].left_hand_landmarks.landmark[:]))#Get number of right hand points
    # numRightHandPoints = len(np.max(session.mediaPipeData[0][:].right_hand_landmarks.landmark[:]))#Get number of left hand points
    numBodyPoints = 33
    numFacePoints = 468
    numHandPoints = 21

    numTrackedPoints = (
        numBodyPoints + numHandPoints * 2 + numFacePoints
    )  # Get total points

    # Create  array of nans the size of number of cams, frame, points, XYC
    mediaPipeData_nCams_nFrames_nImgPts_XYC = np.empty(
        (int(numCams), int(numFrames), int(numTrackedPoints), 3)
    )  # create empty array
    mediaPipeData_nCams_nFrames_nImgPts_XYC[:] = np.NaN  # Fill it with NaNs!

    for camNum in range(numCams):  # Loop through each camera
        for frNum in range(numFrames):  # Loop through each frame
            # make empty arrays for thisFrame's data
            # thisFrame_X = np.empty(numTrackedPoints)
            # thisFrame_X[:] = np.nan
            # thisFrame_Y = thisFrame_X.copy()
            # thisFrame_C = thisFrame_X.copy()

            thisFrame_X_body = np.empty(numBodyPoints)
            thisFrame_X_body[:] = np.nan
            thisFrame_Y_body = thisFrame_X_body.copy()
            thisFrame_C_body = thisFrame_X_body.copy()

            thisFrame_X_face = np.empty(numFacePoints)
            thisFrame_X_face[:] = np.nan
            thisFrame_Y_face = thisFrame_X_face.copy()
            thisFrame_C_face = thisFrame_X_face.copy()

            thisFrame_hands = np.empty(numHandPoints)
            thisFrame_hands[:]= np.nan

            thisFrame_X_lefthand = thisFrame_hands.copy()
            thisFrame_Y_lefthand = thisFrame_hands.copy()
            thisFrame_C_lefthand = thisFrame_hands.copy()

            thisFrame_X_righthand = thisFrame_hands.copy()
            thisFrame_Y_righthand = thisFrame_hands.copy()
            thisFrame_C_righthand = thisFrame_hands.copy()

            fullFrame = True
            if frNum == 107:
                f = 9
            try:
                # pull out ThisFrame's mediapipe data (`mpData.pose_landmarks.landmark` returns something iterable ¯\_(ツ)_/¯)
                thisFrame_poseDataLandMarks = session.mediaPipeData[camNum][
                    frNum
                ].pose_landmarks.landmark  # body ('pose') data
                # stuff body data into pre-allocated nan array
                thisFrame_X_body[:numBodyPoints] = [
                    pp.x for pp in thisFrame_poseDataLandMarks
                ]  # PoseX data - Normalized screen coords (in range [0, 1]) - need multiply by image resultion for pixels
                thisFrame_Y_body[:numBodyPoints] = [
                    pp.y for pp in thisFrame_poseDataLandMarks
                ]  # PoseY data
                thisFrame_C_body[:numBodyPoints] = [
                    pp.visibility for pp in thisFrame_poseDataLandMarks
                ]  #'visibility' is MediaPose's 'confidence' measure in range [0,1]
            except:
                fullFrame = False

            # Right hand data
            try:
                thisFrame_rHandDataLandMarks = session.mediaPipeData[camNum][
                    frNum
                ].right_hand_landmarks.landmark  # right hand data
                thisFrame_X_righthand[:numHandPoints] = [
                    pp.x for pp in thisFrame_rHandDataLandMarks
                ]  # PoseX data - Normalized screen coords (in range [0, 1]) - need multiply by image resultion for pixels
                thisFrame_Y_righthand[:numHandPoints] = [
                    pp.y for pp in thisFrame_rHandDataLandMarks
                ]  # PoseY data
                thisFrame_C_righthand[:numHandPoints] = [
                    pp.visibility for pp in thisFrame_rHandDataLandMarks
                ]  #'visibility' is MediaPose's 'confidence' measure in range [0,1]
            except:
                fullFrame = False

            # Left hand data
            try:
                thisFrame_lHandDataLandMarks = session.mediaPipeData[camNum][
                    frNum
                ].left_hand_landmarks.landmark  # left hand data
                thisFrame_X_lefthand[:numHandPoints ] = [
                    pp.x for pp in thisFrame_lHandDataLandMarks
                ]  # PoseX data - Normalized screen coords (in range [0, 1]) - need multiply by image resultion for pixels
                thisFrame_Y_lefthand[:numHandPoints] = [
                    pp.y for pp in thisFrame_lHandDataLandMarks
                ]  # PoseY data
                thisFrame_C_lefthand[:numHandPoints] = [
                    pp.visibility for pp in thisFrame_lHandDataLandMarks
                ]  #'visibility' is MediaPose's 'confidence' measure in range [0,1]
            except:
                fullFrame = False

            # FaceMeshData
            try:
                thisFrame_faceDataLandMarks = session.mediaPipeData[camNum][
                    frNum
                ].face_landmarks.landmark  # face (mesh) data
                thisFrame_X_face[:numFacePoints] = [
                    pp.x for pp in thisFrame_faceDataLandMarks
                ]  # PoseX data - Normalized screen coords (in range [0, 1]) - need multiply by image resultion for pixels
                thisFrame_Y_face[:numFacePoints] = [
                    pp.y for pp in thisFrame_faceDataLandMarks
                ]  # PoseY data
                # NOTE - There's also Z data in here
                thisFrame_C_face[:numFacePoints] = [
                    pp.visibility for pp in thisFrame_faceDataLandMarks
                ]  #'visibility' is MediaPose's 'confidence' measure in range [0,1]
            except:
                fullFrame = False

            if fullFrame:
                f = 9

            thisFrame_X = np.concatenate((thisFrame_X_body,thisFrame_X_righthand,thisFrame_X_lefthand,thisFrame_X_face))
            thisFrame_Y = np.concatenate((thisFrame_Y_body,thisFrame_Y_righthand,thisFrame_Y_lefthand,thisFrame_Y_face))
            thisFrame_C = np.concatenate((thisFrame_C_body,thisFrame_C_righthand,thisFrame_C_lefthand,thisFrame_C_face))
            # stuff this frame's data into pre-allocated mediaPipeData_.... array
            mediaPipeData_nCams_nFrames_nImgPts_XYC[camNum, frNum, :, 0] = thisFrame_X
            mediaPipeData_nCams_nFrames_nImgPts_XYC[camNum, frNum, :, 1] = thisFrame_Y
            mediaPipeData_nCams_nFrames_nImgPts_XYC[camNum, frNum, :, 2] = thisFrame_C


    # convert from normalized screen coordinates to pixel coordinates
    for camera in range(numCams):
        mediaPipeData_nCams_nFrames_nImgPts_XYC[camera, :, :, 0] *= session.eachCameraResolution['Width'][camera] 
        mediaPipeData_nCams_nFrames_nImgPts_XYC[camera, :, :, 1] *= session.eachCameraResolution['Height'][camera] 



    mediaPipeData_nCams_nFrames_nImgPts_XYC[:, :, 34:, 2] = 1

    np.save(session.dataArrayPath / "mediaPipeData_2d.npy", mediaPipeData_nCams_nFrames_nImgPts_XYC,)

    return mediaPipeData_nCams_nFrames_nImgPts_XYC
    
# #def parseMediaPipe2(session):

#         #numCams = len(session.mediaPipeData) #Get number of cameras
#         #numFrames = len(session.mediaPipeData[0]) #Get number of frames
#         #numBodyPoints = len(np.max(session.mediaPipeData[0][:].pose_landmarks.landmark[:]))#Get number of body points
#         #numFacePoints = len(np.max(session.mediaPipeData[0][:].face_landmarks.landmark[:]))#Get number of face points        
#         #numLeftHandPoints = len(np.max(session.mediaPipeData[0][:].left_hand_landmarks.landmark[:]))#Get number of right hand points
#         #numRightHandPoints = len(np.max(session.mediaPipeData[0][:].right_hand_landmarks.landmark[:]))#Get number of left hand points
#         numBodyPoints = 33
#         numFacePoints = 468
#         numLeftHandPoints = 21
#         numRightHandPoints = 21
        
#         numPoints = numBodyPoints+numFacePoints+numLeftHandPoints+numRightHandPoints #Get total points
#         mediaPipe_nCams_nFrames_nImgPts_XYC = np.ndarray((int(session.numCams),int(session.numFrames),int(numPoints),3)) #Create empty array the size of number of cams, frame, points, XYC
#         for nn in range(session.numCams):#Loop through each camera
#             for ii in range(session.numFrames): #Loop through each frame 
#                 for jj in range(numBodyPoints):
#                     if  session.mediaPipeData[0][ii].pose_landmarks is None: #If that point is not detected
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj,0] = np.nan #Add nan value to that index
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj,1] = np.nan #Add nan value to that index
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj,2] = np.nan #Add nan value to that index
#                     else:
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj,0] = session.mediaPipeData[0][ii].pose_landmarks.landmark[jj].x#Take x pos of that point
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj,1] = session.mediaPipeData[0][ii].pose_landmarks.landmark[jj].y#Take y pos of that point
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj,2] = session.mediaPipeData[0][ii].pose_landmarks.landmark[jj].visibility#Take visibility(confidence) pos of that point
#                 for jj in range(numFacePoints): 
#                     if  session.mediaPipeData[0][ii].face_landmarks is None: #If that point is not detected
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints,0] = np.nan #Add nan value to that index
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints,1] = np.nan #Add nan value to that index
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints,2] = np.nan #Add nan value to that index
#                     else:
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints,0] = session.mediaPipeData[0][ii].face_landmarks.landmark[jj].x#Take x pos of that point
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints,1] = session.mediaPipeData[0][ii].face_landmarks.landmark[jj].y#Take y pos of that point
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints,2] = session.mediaPipeData[0][ii].face_landmarks.landmark[jj].visibility#Take visibility(confidence) pos of that point
#                 for jj in range(numRightHandPoints): 
#                     if  session.mediaPipeData[0][ii].right_hand_landmarks is None: #If that point is not detected
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints,0] = np.nan #Add nan value to that index
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints,1] = np.nan #Add nan value to that index
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints,2] = np.nan #Add nan value to that index
#                     else:
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints,0] = session.mediaPipeData[0][ii].right_hand_landmarks.landmark[jj].x#Take x pos of that point
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints,1] = session.mediaPipeData[0][ii].right_hand_landmarks.landmark[jj].y#Take y pos of that point
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints,2] = session.mediaPipeData[0][ii].right_hand_landmarks.landmark[jj].visibility#Take visibility(confidence) pos of that point
#                 for jj in range(numLeftHandPoints): 
#                     if  session.mediaPipeData[0][ii].left_hand_landmarks is None: #If that point is not detected
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints+numRightHandPoints,0] = np.nan #Add nan value to that index
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints+numRightHandPoints,1] = np.nan #Add nan value to that index
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints+numRightHandPoints,2] = np.nan #Add nan value to that index
#                     else:
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints+numRightHandPoints,0] = session.mediaPipeData[0][ii].left_hand_landmarks.landmark[jj].x#Take x pos of that point
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints+numRightHandPoints,1] = session.mediaPipeData[0][ii].left_hand_landmarks.landmark[jj].y#Take y pos of that point
#                         mediaPipe_nCams_nFrames_nImgPts_XYC[nn,ii,jj+numBodyPoints+numFacePoints+numRightHandPoints,2] = session.mediaPipeData[0][ii].left_hand_landmarks.landmark[jj].visibility#Take visibility(confidence) pos of that point
#             #print(session.mediaPipe_datalist[0].pose_landmarks.landmark[0].x)

#         #path_to_mediapipe_2d = session.dataArrayPath/'mediaPipeData_nCams_nFrames_nImgPts_XY.npy'
        
#         mediaPipeData_nCams_nFrames_nImgPts_XY =  mediaPipe_nCams_nFrames_nImgPts_XYC[:,:,:,0:2].copy()

#         #mediaPipe_nCams_nFrames_nImgPts_XYC[:, :, :, 0] *= 640
#         #mediaPipe_nCams_nFrames_nImgPts_XYC[:, :, :, 1] *= 480
        
#         np.save(session.dataArrayPath / "mediaPipe_2d.npy", mediaPipe_nCams_nFrames_nImgPts_XYC,)

#         return mediaPipe_nCams_nFrames_nImgPts_XYC




# Single Responsibility

# Open a stream between the computer and a camera?
# Open a connection to a usb attached webcam using OpenCV
# Feeds frames to someone else
class Camera:

    def __init__(self):
        pass




import cv2

cap = cv2.VideoCapture(0, cv2.CAP_DSHOW)
# cap = cv2.VideoCapture(2, cv2.CAP_ANY)

cap.set(cv2.CAP_PROP_EXPOSURE, -7)

cap.set(cv2.CAP_PROP_FRAME_WIDTH, 1920)
cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 10800)

cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')) 

while True:
    success, image = cap.read()
    cv2.imshow('blah', image)
    exit_key = cv2.waitKey(1)
    if exit_key ==27:
        break

cap.release()
cv2.destroyAllWindows()




#%%
from pathlib import Path
import freemocap as fmc
import datetime
import logging
from rich import inspect
from rich.console import Console
console = Console()
import shutil
import pandas as pd
#%%



import socket
this_computer_name = socket.gethostname()

freemocap_data_path=None

if this_computer_name=='jon-hallway-XPS-8930':
    freemocap_data_path = Path('/home/jon/Dropbox/FreeMoCapProject/FreeMocap_Data')

elif this_computer_name == 'DESKTOP-DCG6K4F':
    freemocap_data_path = Path(r'C:\Users\jonma\Dropbox\FreeMoCapProject\FreeMocap_Data')
elif this_computer_name == 'DESKTOP-V3D343U':
    freemocap_data_path = Path(r'C:\Users\WindowsPC_Hallway\Dropbox\FreeMoCapProject\FreeMocap_Data')
elif this_computer_name == 'Jons-MacBook-Pro.local':
    freemocap_data_path = Path('/Users/jon/Dropbox/FreeMoCapProject\FreeMocap_Data')

#%%

batch_process_ID = datetime.datetime.now().strftime("batch_process_%Y-%m-%d_%H_%M_%S")
path_to_log_file = freemocap_data_path / (batch_process_ID+'_log.txt')

logging.basicConfig(filename=str(path_to_log_file), filemode='w', format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=logging.INFO)
logging.info('Log file create :D')

#%%
dataframe_init_dict = {
                        'viable_folder_paths':[],
                        'outVid_exists_bool':[]}
viable_folders_dataframe = pd.DataFrame(dataframe_init_dict)
out_vid_path = freemocap_data_path / 'output_animation_videos'
out_vid_path.mkdir(exist_ok=True)

with console.status('Looking for viable folders (that contain synced vids)'):
    for this_folder in freemocap_data_path.glob('**/*'):
        if this_folder.is_dir():
            #check if there is a synced video folder in here
            for this_sub_folder in this_folder.iterdir():
                if this_sub_folder.is_dir():
                    if this_sub_folder.name.lower()[0:4]=='sync':
                        logging.info('Viable Folder found at {}'.format(str(this_folder)))
                        out_vid_exists = False
                        for this_mp4 in this_folder.glob('*.mp4'):
                            out_vid_exists = True
                            shutil.copy(this_mp4, str(out_vid_path))
                            logging.info('OutVid - {} - Copied to - {}'.format(str(this_mp4), str(out_vid_path)))
                        viable_folders_dataframe.loc[len(viable_folders_dataframe.index)] = [str(this_folder), out_vid_exists]


# %%
viable_folders_dataframe


from fastapi import APIRouter
from freemocap.prod.cam.cam_detection import DetectPossibleCameras

# Make a router
camera_router = APIRouter()


# create an endpoint
@camera_router.get("/camera/detect")
def camera_detect():
    dpc = DetectPossibleCameras()
    return dpc.find_available_cameras()



import threading
import cv2
import imutils
import os
import platform

import mediapipe as mp

mp_drawing = mp.solutions.drawing_utils
mp_drawing_styles = mp.solutions.drawing_styles
mp_holistic = mp.solutions.holistic

class VideoSetup(threading.Thread):
    """
    Class to run and thread webcams for preview purposes
    """
    def __init__(self, camID, parameterDictionary, rotNum):
        self.camID = camID
        self.parameterDictionary = parameterDictionary
        self.rotNum = rotNum
        threading.Thread.__init__(self)

    def run(self):
        # print("Starting " + self.previewName)
        self.record(self.parameterDictionary, self.rotNum)

    def record(self, parameterDictionary, rotNum):
        exposure = parameterDictionary.get("exposure")
        resWidth = parameterDictionary.get("resWidth")
        resHeight = parameterDictionary.get("resHeight")
        camWindowName = "Camera" + str(self.camID)+' Preview - Press ESC to exit Setup'
        cv2.namedWindow(camWindowName)

        if platform.system() == 'Windows':
            cap = cv2.VideoCapture(self.camID, cv2.CAP_DSHOW)
        else:
            cap = cv2.VideoCapture(self.camID, cv2.CAP_ANY)


        cap.set(cv2.CAP_PROP_FRAME_WIDTH, resWidth)
        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, resHeight)
        cap.set(cv2.CAP_PROP_EXPOSURE, exposure)
        cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')) 

        # showing values of the properties
        print("__________________________________________")
        print("cv2::videocapture properties for Camera# {}".format(self.camID))
        print("CV_CAP_PROP_FRAME_WIDTH: '{}'".format(cap.get(cv2.CAP_PROP_FRAME_WIDTH)))
        print("CV_CAP_PROP_FRAME_HEIGHT : '{}'".format(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        print("CAP_PROP_FPS : '{}'".format(cap.get(cv2.CAP_PROP_FPS)))
        print("CAP_PROP_EXPOSURE : '{}'".format(cap.get(cv2.CAP_PROP_EXPOSURE)))
        print("CAP_PROP_POS_MSEC : '{}'".format(cap.get(cv2.CAP_PROP_POS_MSEC)))
        print("CAP_PROP_FRAME_COUNT  : '{}'".format(cap.get(cv2.CAP_PROP_FRAME_COUNT)))
        print("CAP_PROP_BRIGHTNESS : '{}'".format(cap.get(cv2.CAP_PROP_BRIGHTNESS)))
        print("CAP_PROP_CONTRAST : '{}'".format(cap.get(cv2.CAP_PROP_CONTRAST)))
        print("CAP_PROP_SATURATION : '{}'".format(cap.get(cv2.CAP_PROP_SATURATION)))
        print("CAP_PROP_HUE : '{}'".format(cap.get(cv2.CAP_PROP_HUE)))
        print("CAP_PROP_GAIN  : '{}'".format(cap.get(cv2.CAP_PROP_GAIN)))
        print("CAP_PROP_CONVERT_RGB : '{}'".format(cap.get(cv2.CAP_PROP_CONVERT_RGB)))
        print("__________________________________________")

        while True:
            ret1, frame1 = cap.read()
            if ret1 == True:
                if rotNum is not None:
                    frame1 = imutils.rotate_bound(frame1, angle=rotNum)
                
                cv2.imshow(camWindowName, frame1)
                if cv2.waitKey(1) & 0xFF == 27:
                    # == ord('q') for q
                    break

            else:
                break
        cv2.destroyWindow(camWindowName)

class MediaPipeVideoSetup(threading.Thread):
    """
    Class to run and thread webcams for preview purposes
    """
    def __init__(self, camID, parameterDictionary, rotNum):
        self.camID = camID
        self.parameterDictionary = parameterDictionary
        self.rotNum = rotNum
        threading.Thread.__init__(self)

    def run(self):
        # print("Starting " + self.previewName)
        self.record(self.parameterDictionary, self.rotNum)

    def record(self, parameterDictionary, rotNum):
        exposure = parameterDictionary.get("exposure")
        resWidth = parameterDictionary.get("resWidth")
        resHeight = parameterDictionary.get("resHeight")
        camWindowName = "RECORDING - Camera" + str(self.camID)+' - Press ESC to exit'

        cv2.namedWindow(camWindowName)

        if platform.system() == 'Windows':
            cap = cv2.VideoCapture(self.camID, cv2.CAP_DSHOW)
        else:
            cap = cv2.VideoCapture(self.camID, cv2.CAP_ANY)

        with mp_holistic.Holistic(
            static_image_mode=True,
            model_complexity=0,
            enable_segmentation=True) as holistic:

            cap.set(cv2.CAP_PROP_FRAME_WIDTH, resWidth)
            cap.set(cv2.CAP_PROP_FRAME_HEIGHT, resHeight)
            cap.set(cv2.CAP_PROP_EXPOSURE, exposure)
            cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')) 

                # showing values of the properties
            print("__________________________________________")
            print("cv2::videocapture properties for Camera# {}".format(self.camID))
            print("CV_CAP_PROP_FRAME_WIDTH: '{}'".format(cap.get(cv2.CAP_PROP_FRAME_WIDTH)))
            print("CV_CAP_PROP_FRAME_HEIGHT : '{}'".format(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)))
            print("CAP_PROP_FPS : '{}'".format(cap.get(cv2.CAP_PROP_FPS)))
            print("CAP_PROP_EXPOSURE : '{}'".format(cap.get(cv2.CAP_PROP_EXPOSURE)))
            print("CAP_PROP_POS_MSEC : '{}'".format(cap.get(cv2.CAP_PROP_POS_MSEC)))
            print("CAP_PROP_FRAME_COUNT  : '{}'".format(cap.get(cv2.CAP_PROP_FRAME_COUNT)))
            print("CAP_PROP_BRIGHTNESS : '{}'".format(cap.get(cv2.CAP_PROP_BRIGHTNESS)))
            print("CAP_PROP_CONTRAST : '{}'".format(cap.get(cv2.CAP_PROP_CONTRAST)))
            print("CAP_PROP_SATURATION : '{}'".format(cap.get(cv2.CAP_PROP_SATURATION)))
            print("CAP_PROP_HUE : '{}'".format(cap.get(cv2.CAP_PROP_HUE)))
            print("CAP_PROP_GAIN  : '{}'".format(cap.get(cv2.CAP_PROP_GAIN)))
            print("CAP_PROP_CONVERT_RGB : '{}'".format(cap.get(cv2.CAP_PROP_CONVERT_RGB)))
            print("__________________________________________")

            while True:
                ret1, frame1 = cap.read()
                if ret1 == True:
                    frame1.flags.writeable = False
                    results = holistic.process(cv2.cvtColor(frame1, cv2.COLOR_BGR2RGB))
                    mp_drawing.draw_landmarks(
                        frame1,
                        results.face_landmarks,
                        mp_holistic.FACEMESH_CONTOURS,
                        landmark_drawing_spec=None,
                        connection_drawing_spec=mp_drawing_styles
                        .get_default_face_mesh_contours_style())
                    mp_drawing.draw_landmarks(
                        frame1,
                        results.pose_landmarks,
                        mp_holistic.POSE_CONNECTIONS,
                        landmark_drawing_spec=mp_drawing_styles
                        .get_default_pose_landmarks_style())
                    if rotNum is not None:
                        frame1 = imutils.rotate_bound(frame1, angle=rotNum)
                    cv2.imshow(camWindowName, frame1)
                    if cv2.waitKey(1) & 0xFF == 27:
                        # == ord('q') for q
                        break

                else:
                    break
            cv2.destroyWindow(camWindowName)



def RunSetup(cam_inputs, rotation_input, paramDict,mediaPipeOverlay):
    """
    Start video setup by threading instances of the VideoSetup class
    """
    if not cam_inputs:
        raise ValueError("Camera input list (cam_inputs) is empty")

    ulist = []

    for cam_input, cam_rotation in zip(cam_inputs, rotation_input):
        if mediaPipeOverlay == True:
            u = MediaPipeVideoSetup(cam_input, paramDict, cam_rotation)
        else:
            u = VideoSetup(cam_input, paramDict, cam_rotation)
        u.start()
        ulist.append(u)

    for k in ulist:
        k.join()



import os
import bpy
import sys
import typing
import inspect
import pkgutil
import importlib
from pathlib import Path

__all__ = (
    "init",
    "register",
    "unregister",
)

blender_version = bpy.app.version

modules = None
ordered_classes = None

def init():
    global modules
    global ordered_classes

    modules = get_all_submodules(Path(__file__).parent)
    ordered_classes = get_ordered_classes_to_register(modules)

def register():
    for cls in ordered_classes:
        bpy.utils.register_class(cls)

    for module in modules:
        if module.__name__ == __name__:
            continue
        if hasattr(module, "register"):
            print(f'registering module: {module.__name__}')
            module.register()

def unregister():
    for cls in reversed(ordered_classes):
        bpy.utils.unregister_class(cls)

    for module in modules:
        if module.__name__ == __name__:
            continue
        if hasattr(module, "unregister"):
            module.unregister()


# Import modules
#################################################

def get_all_submodules(directory):
    return list(iter_submodules(directory, directory.name))

def iter_submodules(path, package_name):
    for name in sorted(iter_submodule_names(path)):
        yield importlib.import_module("." + name, package_name)

def iter_submodule_names(path, root=""):
    for _, module_name, is_package in pkgutil.iter_modules([str(path)]):
        if is_package:
            sub_path = path / module_name
            sub_root = root + module_name + "."
            yield from iter_submodule_names(sub_path, sub_root)
        else:
            yield root + module_name


# Find classes to register
#################################################

def get_ordered_classes_to_register(modules):
    return toposort(get_register_deps_dict(modules))

def get_register_deps_dict(modules):
    my_classes = set(iter_my_classes(modules))
    my_classes_by_idname = {cls.bl_idname : cls for cls in my_classes if hasattr(cls, "bl_idname")}

    deps_dict = {}
    for cls in my_classes:
        deps_dict[cls] = set(iter_my_register_deps(cls, my_classes, my_classes_by_idname))
    return deps_dict

def iter_my_register_deps(cls, my_classes, my_classes_by_idname):
    yield from iter_my_deps_from_annotations(cls, my_classes)
    yield from iter_my_deps_from_parent_id(cls, my_classes_by_idname)

def iter_my_deps_from_annotations(cls, my_classes):
    for value in typing.get_type_hints(cls, {}, {}).values():
        dependency = get_dependency_from_annotation(value)
        if dependency is not None:
            if dependency in my_classes:
                yield dependency

def get_dependency_from_annotation(value):
    if blender_version >= (2, 93):
        if isinstance(value, bpy.props._PropertyDeferred):
            return value.keywords.get("type")
    else:
        if isinstance(value, tuple) and len(value) == 2:
            if value[0] in (bpy.props.PointerProperty, bpy.props.CollectionProperty):
                return value[1]["type"]
    return None

def iter_my_deps_from_parent_id(cls, my_classes_by_idname):
    if bpy.types.Panel in cls.__bases__:
        parent_idname = getattr(cls, "bl_parent_id", None)
        if parent_idname is not None:
            parent_cls = my_classes_by_idname.get(parent_idname)
            if parent_cls is not None:
                yield parent_cls

def iter_my_classes(modules):
    base_types = get_register_base_types()
    for cls in get_classes_in_modules(modules):
        if any(base in base_types for base in cls.__bases__):
            if not getattr(cls, "is_registered", False):
                yield cls

def get_classes_in_modules(modules):
    classes = set()
    for module in modules:
        for cls in iter_classes_in_module(module):
            classes.add(cls)
    return classes

def iter_classes_in_module(module):
    for value in module.__dict__.values():
        if inspect.isclass(value):
            yield value

def get_register_base_types():
    return set(getattr(bpy.types, name) for name in [
        "Panel", "Operator", "PropertyGroup",
        "AddonPreferences", "Header", "Menu",
        "Node", "NodeSocket", "NodeTree",
        "UIList", "RenderEngine",
        "Gizmo", "GizmoGroup",
    ])


# Find order to register to solve dependencies
#################################################

def toposort(deps_dict):
    sorted_list = []
    sorted_values = set()
    while len(deps_dict) > 0:
        unsorted = []
        for value, deps in deps_dict.items():
            if len(deps) == 0:
                sorted_list.append(value)
                sorted_values.add(value)
            else:
                unsorted.append(value)
        deps_dict = {value : deps_dict[value] - sorted_values for value in unsorted}
    return sorted_list



from tqdm import tqdm
import cv2
import os
import platform

def TestDevice(source):
    """
    Checks to see if there is a camera available at the input
    """
    if platform.system() == 'Windows':
        cap = cv2.VideoCapture(source, cv2.CAP_DSHOW)
    else:
        cap = cv2.VideoCapture(source, cv2.CAP_ANY)
        
    # if cap is None or not cap.isOpened():
    # print('Warning: unable to open video source: ', source)

    if cap.isOpened():
        success, image = cap.read()
        if success:
            # print('Opened: ',source)
            # print('Exposure: '+ str(cap.get(cv2.CAP_PROP_EXPOSURE)))
            # time.sleep(3)
            cap.release()
            cv2.destroyAllWindows()
            open_cam = source
            return open_cam
        else:
            return None
    else:
        return None


def CreateAvailableCamList():
    """
    Loops through the first 20 ports (this number is overkill, really) and creates a list of ports that have a camera available on them 
    """
    openCamList = []
    for x in tqdm(range(20)):  # range 20 right now to be safe
        openCamera = TestDevice(x)
        if openCamera is not None:
            openCamList.append(openCamera)

    return openCamList



import freemocap as fmc

from FMC_MultiCamera import FMC_MultiCamera

from pathos.helpers import mp as pathos_mp_helper

if __name__ =='__main__':
    pathos_mp_helper.freeze_support()

    multi_cam =  FMC_MultiCamera(freemocap_data_folder='C:/Users/WindowsPC_Hallway/Dropbox/FreeMoCapProject/FreeMocap_Data')
    sessionID = multi_cam._rec_name
    multi_cam.start(standalone_mode=True)    
    f=9



import numpy as np
import bpy
from mathutils import Vector
import math
import sys
from pathlib import Path
import csv
from math import degrees


# get arguments
argv = sys.argv
argv = argv[argv.index("--") + 1:] 
 
# get npy session data
input_frames_data_npy = argv[0]

# get session path
sesh_path = argv[1]

#get session ID
sesh_ID = argv[2]

# A clean array of skeleton data where every marker was tracked that we can build off of
clean_skeleton_position_arr = Path(__file__).parent.resolve() / "build_data.npy"

#3D array holding [[[x, y, z], [x, y, z]], [[x, y, z], [x, y, z]]] 
mediapipe_frames_arr = np.load(input_frames_data_npy)

# Saved clean frame where all points are visible and not NANs 
markers_list = np.load(clean_skeleton_position_arr)
#a list containing empty objects
order_of_markers = []

#frame rate 
bpy.context.scene.render.fps = 30

#dictionary corresponding to marker numbers and body parts
mediapipe_tracked_point_names = {
    0:"nose",
    1:"left_eye_inner",
    2:"left_eye",
    3:"left_eye_outer",
    4:"right_eye_inner",
    5:"right_eye",
    6:"right_eye_outer",
    7:"left_ear",
    8:"right_ear",
    9:"mouth_left",
    10:"mouth_right",
    11:"shoulder_L", 
    12:"shoulder_R", 
    13:"elbow_L", 
    14:"elbow_R", 
    23:"hip_L",
    24:"hip_R", 
    25:"knee_L", 
    26:"knee_R",
    27:"ankle_L",
    28:"ankle_R",
    29:"heel_L", 
    30:"heel_R",
    31:"toe_L",
    32:"toe_R" }

#-----------------------------------------------------------------------------------   

# set project unit 
bpy.context.scene.unit_settings.length_unit = 'METERS'

#create sub-collection of empties
cname = "mediapipe_empty_objects"
empty_collection = bpy.data.collections.new(cname)
bpy.data.collections.get("Collection").children.link(empty_collection)
master_collection = bpy.data.collections.get("Collection")


#iterate through arr and create an empty object at that location for each element
for index, col in enumerate(markers_list):
    # parse string float value into floats, create Vector, set position to Vector

    if math.isnan(col[0]):
        col[0] = np.nan
    if math.isnan(col[1]):
        col[1] = np.nan
    if math.isnan(col[2]):
        col[2] = np.nan
    coord = Vector(((float(col[0])*0.001), (float(col[1])*0.001), (float(col[2]))* 0.001))

        
    #empties
    bpy.ops.object.add(type='EMPTY', location=coord)  
    this_empty = bpy.context.active_object  
    this_empty.name = "empty_" + str(index)
    
    if index in mediapipe_tracked_point_names.keys():
        this_empty.name += "_" + str(mediapipe_tracked_point_names[index])
    else:
        if index > 32 and index < 54:
            this_empty.name += "_" + str("right_hand")
        elif index >= 54 and index < 76:
            this_empty.name += "_" + str("left_hand")
        else:
            this_empty.name += "_" + str("face")
            
    order_of_markers.append(this_empty)

    #link this empty to the scene's Empty collection
    empty_collection.objects.link( this_empty )
    #unlink empty from master collection
    master_collection.objects.unlink(this_empty) 
    #set location 
    this_empty.location = coord
    #set the display size of the empty
    this_empty.empty_display_size = 0.02
    
#--------------------------------------------------------------
#Virtual Markers!

# marker relationships:

# - "weight": the weighted average of multiple markers. the virtual_markers[x] contains the 
#list of markers that affect this virtual one. the weights[x] contains their corresponding weights in order.

bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

#Create virtual marker where parameters are the name of the marker, the markers that affect its position, and each of their weights.
def create_marker_weight(name, markers, weighted):
    center = Vector((0, 0, 0))
    weight_iter = 0
    for x in markers:
        center += x.location*weighted[weight_iter]
        weight_iter += 1
    coord = Vector((float(center[0]), float(center[1]), float(center[2])))
    bpy.ops.object.add(type='EMPTY', location=coord)
    this_empty = bpy.context.active_object  
    this_empty.name = name
    #link this empty to the scene's Empty collection
    empty_collection.objects.link( this_empty )
    #unlink empty from master collection
    master_collection.objects.unlink(this_empty) 
    this_empty.location = coord
    this_empty.empty_display_size = 0.02
    virtual_markers.append(this_empty)

#Keeping track of virtual marker info using arrays, where each marker is an index in each array
v_relationship = []
virtual_markers = []
surrounding_markers = []
weights = []

#updates data and creates virtual marker
def update_virtual_data(relationship, surrounding, vweights, vname):
    v_relationship.append(relationship)
    surrounding_markers.append(surrounding)
    weights.append(vweights)
    create_marker_weight(vname, surrounding, vweights)
         
#-----------------------------------------------------------------
#Define relationships and create virtual markers

#Neck Base: Halfway between order_of_markers[12] shoulder_L and order_of_markers[12] shoulder_R
l0 = [order_of_markers[12], order_of_markers[11]]
w0 = [0.5, 0.5]
update_virtual_data("weight", l0, w0, "empty_neck")

#Waist Base: Halfway between order_of_markers[24] and order_of_markers[23] 
l0 = [order_of_markers[24], order_of_markers[23]]
w0 = [0.5, 0.5]
update_virtual_data("weight", l0, w0, "empty_waist")

#Update the location of virtual markers on each frame
def update_virtual_marker(index):
    if(v_relationship[index] == "weight"):
        center = Vector((0, 0, 0))
        weight_iter = 0
        for x in surrounding_markers[index]:
            center += x.location*weights[index][weight_iter]
            weight_iter += 1
        coord = Vector((float(center[0]), float(center[1]), float(center[2])))
    virtual_markers[index].location = coord
    
#-----------------------------------------------------------------------------------
# Bones! 
    
#adds child bone given corresponding parent and empty
#bone tail will appear at the location of empty
def add_child_bone(bone_name, empty_child, empty_parent):
    #Set armature selected
    armature_data.select_set(state=True)
    #Set edit mode
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
    #Create a new bone
    new_bone = armature_data.data.edit_bones.new(bone_name)
    #Set bone's size

    new_bone.head = (np.nan,np.nan,np.nan)
    new_bone.tail = (np.nan,np.nan,np.nan)
    #Set bone's location to wheel
    new_bone.matrix = empty_parent.matrix_world

    #set location of bone head
    new_bone.head =  empty_child.location
    #set location of bone tail
    new_bone.tail = empty_parent.location
    return new_bone

#Create armature object
armature = bpy.data.armatures.new('Armature')
armature.name = "Armature_MediaPipe"
armature_object = bpy.data.objects.new('Armature', armature)
#set name of armature to specify data type (media pipe)
armature_object.name = "Armature_MediaPipe_Object"
#Link armature object to our scene
bpy.context.collection.objects.link(armature_object)
#Make armature variable
armature_data = bpy.data.objects[armature_object.name]
#Set armature active
bpy.context.view_layer.objects.active = armature_data
#Set armature selected
armature_data.select_set(state=True)
#Set edit mode
bpy.ops.object.mode_set(mode='EDIT', toggle=False)
#Set bones in front and show axis
armature_data.show_in_front = True
#True to show axis orientation of bones and false to hide it
armature_data.data.show_axes = False

#get armature object
def get_armature():
    for ob in bpy.data.objects:
        if ob.name == 'Armature_MediaPipe_Object':
            armature = ob
            break
    return armature

armature = get_armature()

#a list of tuples where each element represents the info for one bone, and 
#the tuple contains the bone name, the marker of the head, and the marker of the tail
list_of_bones_order = [('upper_arm_L', order_of_markers[11], order_of_markers[13]),
        ('upper_arm_R', order_of_markers[12], order_of_markers[14]),
        ('lower_arm_L', order_of_markers[13], order_of_markers[15]),
        ('lower_arm_R', order_of_markers[14], order_of_markers[16]),
        ('upper_leg_L', order_of_markers[23], order_of_markers[25]),
        ('upper_leg_R', order_of_markers[24], order_of_markers[26]),
        ('lower_leg_L', order_of_markers[25], order_of_markers[27]),
        ('lower_leg_R', order_of_markers[26], order_of_markers[28]),
        ('heel_L', order_of_markers[27], order_of_markers[29]),
        ('heel_R', order_of_markers[28], order_of_markers[30]),
        ('foot_L', order_of_markers[29], order_of_markers[31]),
        ('foot_R', order_of_markers[30], order_of_markers[32]),
        ('neck', order_of_markers[0], virtual_markers[0]),
        ('eye_L', order_of_markers[0], order_of_markers[2]),
        ('ear_L', order_of_markers[2], order_of_markers[7]),
        ('eye_R', order_of_markers[0], order_of_markers[5]),
        ('ear_R', order_of_markers[5], order_of_markers[8]),
        ('shoulder_L', virtual_markers[0], order_of_markers[11]),
        ('shoulder_R', virtual_markers[0], order_of_markers[12]),
        ('root', virtual_markers[1], virtual_markers[0]),
        ('hip_L', virtual_markers[1], order_of_markers[23]),
        ('hip_R', virtual_markers[1], order_of_markers[24]),
        ]
        
#based on marker # from order_of_markers array add bones for hands:
left_hand_offset = 54
    
left_hand = [('handL0', order_of_markers[0+left_hand_offset], order_of_markers[1+left_hand_offset]),
    ('handL1', order_of_markers[1+left_hand_offset], order_of_markers[2+left_hand_offset]),
    ('handL2', order_of_markers[2+left_hand_offset], order_of_markers[3+left_hand_offset]),
    ('handL3', order_of_markers[3+left_hand_offset], order_of_markers[4+left_hand_offset]),
    ('handL5', order_of_markers[0+left_hand_offset], order_of_markers[5+left_hand_offset]),
    ('handL6', order_of_markers[5+left_hand_offset], order_of_markers[6+left_hand_offset]),
    ('handL7', order_of_markers[6+left_hand_offset], order_of_markers[7+left_hand_offset]),
    ('handL8', order_of_markers[7+left_hand_offset], order_of_markers[8+left_hand_offset]),
    ('handL9', order_of_markers[0+left_hand_offset], order_of_markers[9+left_hand_offset]),
    ('handL10', order_of_markers[9+left_hand_offset], order_of_markers[10+left_hand_offset]),
    ('handL11', order_of_markers[10+left_hand_offset], order_of_markers[11+left_hand_offset]),
    ('handL12', order_of_markers[11+left_hand_offset], order_of_markers[12+left_hand_offset]),
    ('handL13', order_of_markers[0+left_hand_offset], order_of_markers[13+left_hand_offset]),
    ('handL14', order_of_markers[13+left_hand_offset], order_of_markers[14+left_hand_offset]),
    ('handL15', order_of_markers[14+left_hand_offset], order_of_markers[15+left_hand_offset]),
    ('handL16', order_of_markers[15+left_hand_offset], order_of_markers[16+left_hand_offset]),
    ('handL18', order_of_markers[0+left_hand_offset], order_of_markers[17+left_hand_offset]),
    ('handL19', order_of_markers[17+left_hand_offset], order_of_markers[18+left_hand_offset]),
    ('handL20', order_of_markers[18+left_hand_offset], order_of_markers[19+left_hand_offset]),
    ('wristL', order_of_markers[15], order_of_markers[0+left_hand_offset]),
    ('handL21', order_of_markers[19+left_hand_offset], order_of_markers[20+left_hand_offset]),]
    
    
right_hand_offset = 33

right_hand = [('handR0', order_of_markers[18], order_of_markers[1+right_hand_offset]),
    ('handR1', order_of_markers[1+right_hand_offset], order_of_markers[2+right_hand_offset]),
    ('handR2', order_of_markers[2+right_hand_offset], order_of_markers[3+right_hand_offset]),
    ('handR3', order_of_markers[3+right_hand_offset], order_of_markers[4+right_hand_offset]),
    ('handR5', order_of_markers[18], order_of_markers[5+right_hand_offset]),
    ('handR6', order_of_markers[5+right_hand_offset], order_of_markers[6+right_hand_offset]),
    ('handR7', order_of_markers[6+right_hand_offset], order_of_markers[7+right_hand_offset]),
    ('handR8', order_of_markers[7+right_hand_offset], order_of_markers[8+right_hand_offset]),
    ('handR9', order_of_markers[18], order_of_markers[9+right_hand_offset]),
    ('handR10', order_of_markers[9+right_hand_offset], order_of_markers[10+right_hand_offset]),
    ('handR11', order_of_markers[10+right_hand_offset], order_of_markers[11+right_hand_offset]),
    ('handR12', order_of_markers[11+right_hand_offset], order_of_markers[12+right_hand_offset]),
    ('handR13', order_of_markers[18], order_of_markers[13+right_hand_offset]),
    ('handR14', order_of_markers[13+right_hand_offset], order_of_markers[14+right_hand_offset]),
    ('handR15', order_of_markers[14+right_hand_offset], order_of_markers[15+right_hand_offset]),
    ('handR16', order_of_markers[15+right_hand_offset], order_of_markers[16+right_hand_offset]),
    ('handR18', order_of_markers[18], order_of_markers[17+right_hand_offset]),
    ('handR19', order_of_markers[17+right_hand_offset], order_of_markers[18+right_hand_offset]),
    ('handR20', order_of_markers[18+right_hand_offset], order_of_markers[19+right_hand_offset]),
    ('wristR', order_of_markers[16], order_of_markers[18]),
    ('handR21', order_of_markers[19+right_hand_offset], order_of_markers[20+right_hand_offset]),]

#helper to create armature from list of tuples
def tuple_to_armature(bones):
    for bone_name, bone_head, bone_tail in bones:
        add_child_bone(bone_name, bone_head, bone_tail)
        
#Set armature selected
armature_data.select_set(state=True)
#Set edit mode
bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        
#create all bones for skeleton body and hands
tuple_to_armature(list_of_bones_order)
tuple_to_armature(right_hand)
tuple_to_armature(left_hand)

#parent heads and tails to empties
#use bone constraints 
def parent_to_empties(bone_name, head, tail):
    #enter pose mode
    bpy.ops.object.posemode_toggle()
    marker = armature.data.bones[bone_name]
    #Set marker selected
    marker.select = True
    #Set marker active
    bpy.context.object.data.bones.active = marker
    bone = bpy.context.object.pose.bones[bone_name]
    #Copy Location Pose constraint: makes the bone's head follow the given empty
    bpy.ops.pose.constraint_add(type='COPY_LOCATION')
    bone.constraints["Copy Location"].target = head
    #Stretch To Pose constraint: makes the bone's tail follow the given empty
    #stretches the bones to reach the tail to that empty so head location is not affected
    bpy.ops.pose.constraint_add(type='STRETCH_TO')
    bone.constraints["Stretch To"].target = tail
    #exit pose mode
    bpy.ops.object.posemode_toggle()
    
#set parents of heads and tails for each bone 
def tuple_to_parented(bones):
    for bone_name, bone_head, bone_tail in bones:
        parent_to_empties(bone_name, bone_head, bone_tail)

tuple_to_parented(list_of_bones_order)
tuple_to_parented(right_hand)
tuple_to_parented(left_hand)

#-----------------------------------------------------------------------------------
# Animate! 
#find number of frames in animation
num_frames = len(mediapipe_frames_arr)

#change start frame of animation
bpy.context.scene.frame_start = 1
#change end frame of animation
bpy.context.scene.frame_end = num_frames - 1

#create a new handler to change empty positions every frame
def my_handler(scene): 
    bpy.ops.object.mode_set(mode='OBJECT')
    #keep track of current_marker
    current_marker = 0
    #find the current frame number
    frame = scene.frame_current
    #get the list of marker points from the current frame
    markers_list = mediapipe_frames_arr[frame]
    #iterate through list of markers in this frame
    for col in markers_list:
        frame = scene.frame_current

        if math.isnan(col[0]) and math.isnan(col[1]) and math.isnan(col[2]):
           #marker was not tracked, pass
           nan_frame = True
           pass
        else:
           #Flip y and z coordinates, and negate y to account for different coordinate system
           nan_frame = False
           coord = Vector(((float(col[0])*0.001), (float(col[1])*0.001), (float(col[2]))* 0.001))
           if len(order_of_markers) > 0:
                empty = order_of_markers[current_marker]
                empty.location = coord
                for index in range(len(virtual_markers)):
                    update_virtual_marker(index)
        current_marker += 1 

    
    #keyframe bones
    #Goes through each bone
    if not nan_frame:
        for editBone in get_armature().data.edit_bones:
            boneName = editBone.name
            poseBone = arm.pose.bones[boneName]
            poseBone.keyframe_insert('rotation_euler', frame=scene.frame_current)
            poseBone.keyframe_insert('location', frame=scene.frame_current)
            poseBone.keyframe_insert('scale', frame=scene.frame_current)

#-----------------------------------------------------------------------------------
#script to create a mesh of the armature 
def CreateMesh():
    obj = get_armature()

    if obj == None:
        print( "No selection" )
    elif obj.type != 'ARMATURE':
        print( "Armature expected" )
    else:
        return processArmature( bpy.context, obj )

#Use armature to create base object
def armToMesh( arm ):
    name = arm.name + "_mesh"
    dataMesh = bpy.data.meshes.new( name + "Data" )
    mesh = bpy.data.objects.new( name, dataMesh )
    mesh.matrix_world = arm.matrix_world.copy()
    return mesh

#Make vertices and faces 
def boneGeometry( l1, l2, x, z, baseSize, l1Size, l2Size, base ):
    x1 = x  * baseSize * l1Size * 0.8
    z1 = z  * baseSize * l2Size * 0.8
    
    x2 = Vector( (np.nan, np.nan, np.nan) )
    z2 = Vector( (np.nan, np.nan, np.nan) )

    verts = [
        l1 - x1 + z1,
        l1 + x1 + z1,
        l1 - x1 - z1,
        l1 + x1 - z1,
        l2 - x2 + z2,
        l2 + x2 + z2,
        l2 - x2 - z2,
        l2 + x2 - z2
        ] 

    faces = [
        (base+3, base+1, base+0, base+2),
        (base+6, base+4, base+5, base+7),
        (base+4, base+0, base+1, base+5),
        (base+7, base+3, base+2, base+6),
        (base+5, base+1, base+3, base+7),
        (base+6, base+2, base+0, base+4)
        ]

    return verts, faces

#Process the armature, goes through its bones and creates the mesh
def processArmature(context, arm, genVertexGroups = True):
    print("processing armature {0}".format(arm.name))

    #Creates the mesh object
    meshObj = armToMesh( arm )
    context.collection.objects.link( meshObj )

    verts = []
    edges = []
    faces = []
    vertexGroups = {}

    bpy.ops.object.mode_set(mode='EDIT')

    try:
        #Goes through each bone
        for editBone in get_armature().data.edit_bones:
            boneName = editBone.name
            if boneName == "":
                editBone.name = "bone"
                boneName = "bone"
            poseBone = arm.pose.bones[boneName]

            #Gets edit bone informations
            editBoneHead = editBone.head
            editBoneTail = editBone.tail
            editBoneVector = editBoneTail - editBoneHead
            editBoneSize = editBoneVector.dot( editBoneVector )
            editBoneRoll = editBone.roll
            editBoneX = editBone.x_axis
            editBoneZ = editBone.z_axis
            editBoneHeadRadius = editBone.head_radius
            editBoneTailRadius = editBone.tail_radius

            #Creates the mesh data for the bone
            baseIndex = len(verts)
            baseSize = math.sqrt( editBoneSize )
            newVerts, newFaces = boneGeometry( editBoneHead, editBoneTail, editBoneX, editBoneZ, baseSize, editBoneHeadRadius, editBoneTailRadius, baseIndex )
            verts.extend( newVerts )
            faces.extend( newFaces )

            #Creates the weights for the vertex groups
            vertexGroups[boneName] = [(x, 1.0) for x in range(baseIndex, len(verts))]
        #Assigns the geometry to the mesh
        meshObj.data.from_pydata(verts, edges, faces)

    except:
        bpy.ops.object.mode_set(mode='OBJECT')
    else:
        bpy.ops.object.mode_set(mode='OBJECT')
    #Assigns the vertex groups
    if genVertexGroups:
        for name1, vertexGroup in vertexGroups.items():
            groupObject = meshObj.vertex_groups.new(name = name1)
            for (index, weight) in vertexGroup:
                groupObject.add([index], weight, 'REPLACE')

    #Creates the armature modifier
    modifier = meshObj.modifiers.new('ArmatureMod', 'ARMATURE')
    modifier.object = arm
    modifier.use_bone_envelopes = False
    modifier.use_vertex_groups = True

    meshObj.data.update()

    return meshObj

mesh_obob = CreateMesh()

#-----------------------------------------------------------------------------------
# Clean up the mesh by removing duplicate vertices, make sure all faces are quads, etc

checked = set()
for selected_object in bpy.data.objects:
    if selected_object.type != 'MESH':
        continue
    meshdata = selected_object.data
    if meshdata in checked:
        continue
    else:
        checked.add(meshdata)
    bpy.context.view_layer.objects.active = selected_object
    bpy.ops.object.editmode_toggle()
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.remove_doubles()
    bpy.ops.mesh.tris_convert_to_quads()
    bpy.ops.mesh.normals_make_consistent()
    bpy.ops.object.editmode_toggle()
#Set armature active
bpy.context.view_layer.objects.active = armature_data
#Set armature selected
armature_data.select_set(state=True)
#-----------------------------------------------------------------------------------
#function to register custom handler
bpy.app.handlers.frame_change_post.clear()
def register():
    bpy.app.handlers.frame_change_post.append(my_handler)
   
def unregister():
    bpy.app.handlers.frame_change_post.remove(my_handler)
    
register()

scene = bpy.context.scene

bpy.ops.object.mode_set(mode='OBJECT')

bpy.ops.object.mode_set(mode='EDIT')

arm = get_armature()

#Connected according to: https://google.github.io/mediapipe/solutions/hands.html#hand-landmark-model

arm.data.edit_bones['shoulder_L'].parent = arm.data.edit_bones['root']
arm.data.edit_bones['shoulder_R'].parent = arm.data.edit_bones['root']

arm.data.edit_bones['upper_arm_L'].parent = arm.data.edit_bones['shoulder_L']
arm.data.edit_bones['upper_arm_R'].parent = arm.data.edit_bones['shoulder_R']
arm.data.edit_bones['lower_arm_L'].parent = arm.data.edit_bones['upper_arm_L']
arm.data.edit_bones['lower_arm_R'].parent = arm.data.edit_bones['upper_arm_R']

arm.data.edit_bones['hip_L'].parent = arm.data.edit_bones['root']
arm.data.edit_bones['hip_R'].parent = arm.data.edit_bones['root']
arm.data.edit_bones['upper_leg_L'].parent = arm.data.edit_bones['hip_L']
arm.data.edit_bones['upper_leg_R'].parent = arm.data.edit_bones['hip_R']
arm.data.edit_bones['lower_leg_L'].parent = arm.data.edit_bones['upper_leg_L']
arm.data.edit_bones['lower_leg_R'].parent = arm.data.edit_bones['upper_leg_R']
arm.data.edit_bones['heel_L'].parent = arm.data.edit_bones['lower_leg_L']
arm.data.edit_bones['heel_R'].parent = arm.data.edit_bones['lower_leg_R']
arm.data.edit_bones['foot_L'].parent = arm.data.edit_bones['heel_L']
arm.data.edit_bones['foot_R'].parent = arm.data.edit_bones['heel_R']


arm.data.edit_bones['neck'].parent = arm.data.edit_bones['root']
arm.data.edit_bones['eye_L'].parent = arm.data.edit_bones['neck']
arm.data.edit_bones['eye_R'].parent = arm.data.edit_bones['neck']
arm.data.edit_bones['ear_L'].parent = arm.data.edit_bones['eye_L']
arm.data.edit_bones['ear_R'].parent = arm.data.edit_bones['eye_R']

    
arm.data.edit_bones['wristL'].parent = arm.data.edit_bones['lower_arm_L']
arm.data.edit_bones['wristR'].parent = arm.data.edit_bones['lower_arm_R']

#------------------------------------------------------------------------

arm.data.edit_bones['handR0'].parent = arm.data.edit_bones['wristR']
arm.data.edit_bones['handR1'].parent = arm.data.edit_bones['handR0']
arm.data.edit_bones['handR2'].parent = arm.data.edit_bones['handR1']
arm.data.edit_bones['handR3'].parent = arm.data.edit_bones['handR2']

arm.data.edit_bones['handR5'].parent = arm.data.edit_bones['wristR']
arm.data.edit_bones['handR6'].parent = arm.data.edit_bones['handR5']
arm.data.edit_bones['handR7'].parent = arm.data.edit_bones['handR6']
arm.data.edit_bones['handR8'].parent = arm.data.edit_bones['handR7']

arm.data.edit_bones['handR9'].parent = arm.data.edit_bones['wristR']
arm.data.edit_bones['handR10'].parent = arm.data.edit_bones['handR9']
arm.data.edit_bones['handR11'].parent = arm.data.edit_bones['handR10']
arm.data.edit_bones['handR12'].parent = arm.data.edit_bones['handR11']

arm.data.edit_bones['handR13'].parent = arm.data.edit_bones['wristR']
arm.data.edit_bones['handR14'].parent = arm.data.edit_bones['handR13']
arm.data.edit_bones['handR15'].parent = arm.data.edit_bones['handR14']
arm.data.edit_bones['handR16'].parent = arm.data.edit_bones['handR15']

arm.data.edit_bones['handR18'].parent = arm.data.edit_bones['wristR']
arm.data.edit_bones['handR19'].parent = arm.data.edit_bones['handR18']
arm.data.edit_bones['handR20'].parent = arm.data.edit_bones['handR19']
arm.data.edit_bones['handR21'].parent = arm.data.edit_bones['handR20']

##------------------------------------------------------------------------


arm.data.edit_bones['handL0'].parent = arm.data.edit_bones['wristL']
arm.data.edit_bones['handL1'].parent = arm.data.edit_bones['handL0']
arm.data.edit_bones['handL2'].parent = arm.data.edit_bones['handL1']
arm.data.edit_bones['handL3'].parent = arm.data.edit_bones['handL2']

arm.data.edit_bones['handL5'].parent = arm.data.edit_bones['wristL']
arm.data.edit_bones['handL6'].parent = arm.data.edit_bones['handL5']
arm.data.edit_bones['handL7'].parent = arm.data.edit_bones['handL6']
arm.data.edit_bones['handL8'].parent = arm.data.edit_bones['handL7']

arm.data.edit_bones['handL9'].parent = arm.data.edit_bones['wristL']
arm.data.edit_bones['handL10'].parent = arm.data.edit_bones['handL9']
arm.data.edit_bones['handL11'].parent = arm.data.edit_bones['handL10']
arm.data.edit_bones['handL12'].parent = arm.data.edit_bones['handL11']

arm.data.edit_bones['handL13'].parent = arm.data.edit_bones['wristL']
arm.data.edit_bones['handL14'].parent = arm.data.edit_bones['handL13']
arm.data.edit_bones['handL15'].parent = arm.data.edit_bones['handL14']
arm.data.edit_bones['handL16'].parent = arm.data.edit_bones['handL15']

arm.data.edit_bones['handL18'].parent = arm.data.edit_bones['wristL']
arm.data.edit_bones['handL19'].parent = arm.data.edit_bones['handL18']
arm.data.edit_bones['handL20'].parent = arm.data.edit_bones['handL19']
arm.data.edit_bones['handL21'].parent = arm.data.edit_bones['handL20']

for editBone in get_armature().data.edit_bones:
    # this line is important, when inheriting scale and using the stretch to constraint the bones freak out
    editBone.inherit_scale = 'NONE'


##------------------------------------------------------------------------

#Select objects to export
col = bpy.data.collections.get("Collection")
            
bpy.ops.object.mode_set(mode='OBJECT')

#Run thorugh frames, export a csv file with local bone rotations
# open the file in the write mode
f = open(sesh_path + sesh_ID + '_local_bone_rotations.csv', 'w')

# create the csv writer
writer = csv.writer(f)
writer.writerow(["Frame_number", "Bone name", "X", "Y", "Z"])
for frame in range(scene.frame_start, scene.frame_end):
    scene.frame_set(frame)
    arm = get_armature()
    coords_per_frame = [frame]
    for bone in arm.pose.bones:
        if bone.parent:
            Original_rot = bone.matrix_basis.copy()

            a = bone.matrix.copy()
            b = bone.parent.matrix.copy()

            q1 = a.to_quaternion()

            q2 = b.to_quaternion()

            difQuat = q1.rotation_difference(q2)

            difEuler = difQuat.to_euler()
            mat = difEuler
        else:
            a = bone.matrix.copy()
            q1 = a.to_quaternion()
            mat = q1.to_euler()
        x = degrees(mat.x)
        y = degrees(mat.y)
        z = degrees(mat.z)
        
        if x < 0.0001:
            x = 0.0
        if y < 0.0001:
            y = 0.0
        if z < 0.0001:
            z = 0.0
        coords_per_frame.append(bone.name)
        coords_per_frame.append(x)
        coords_per_frame.append(y)
        coords_per_frame.append(z)
    writer.writerow(coords_per_frame)


bpy.ops.object.mode_set(mode='OBJECT')


#set keyframes for whole animation
for frame in range(scene.frame_start, scene.frame_end):
   scene.frame_set(frame)
   
#Select objects to export
col = bpy.data.collections.get("Collection")
if col:
   for obj in col.objects:
       if "Armature_MediaPipe_Object" == obj.name:
            obj.select_set(True)


#Bake the skeletal animation
# ensure that only the armature is selected in Object mode
bpy.ops.object.mode_set(mode='OBJECT')
bpy.ops.object.select_all(action='DESELECT')

empties = [e for e in bpy.data.objects
        if e.type.startswith('EMPTY')]
for e in empties:
    e.select_set(True)
    
    
#Set armature active
bpy.context.view_layer.objects.active = bpy.data.objects[get_armature().name]
#Set armature selected
bpy.data.objects[get_armature().name].select_set(state=True)

#Change to pose mode
bpy.ops.object.mode_set(mode='POSE')

#Bake the animation
bpy.ops.nla.bake(frame_start=scene.frame_start, frame_end=scene.frame_end, only_selected=False, visual_keying=True, clear_constraints=True, use_current_action=False, bake_types={'POSE', 'OBJECT'})
    
bpy.ops.object.mode_set(mode='OBJECT')

#Select objects to export
col = bpy.data.collections.get("Collection")
if col:
   for obj in col.objects:
       if "Armature" in obj.name:
            obj.select_set(True)
       if obj.name == "Cube" or obj.name == "Light" or obj.name == "Camera":
            bpy.data.objects.remove(obj)
 
for e in empties:
    e.select_set(True)
            
#save blender file
bpy.ops.wm.save_as_mainfile(filepath= sesh_path + sesh_ID + '.blend')


#export FBX
bpy.ops.export_scene.fbx(filepath= sesh_path + sesh_ID + '.fbx', path_mode='RELATIVE', bake_anim=True, use_selection=True, object_types={'MESH', 'ARMATURE', 'EMPTY'}, use_mesh_modifiers = False, add_leaf_bones = False, axis_forward = '-X', axis_up = 'Z', bake_anim_use_all_bones = False, bake_anim_use_nla_strips = False, bake_anim_use_all_actions = False, bake_anim_force_startend_keying = False) 

#export GLTF
bpy.ops.export_scene.gltf(filepath=sesh_path + sesh_ID + '.gltf', export_format='GLTF_EMBEDDED', use_selection=True, ui_tab='ANIMATION')

#export USD
bpy.ops.wm.usd_export(filepath=sesh_path + sesh_ID + '.usdc', selected_objects_only=True, export_animation=True)


#--------------------------------------------------------------------------------------------
#Create a panel to display local bone angles in relation to parent bone
class LOCAL_PT_BoneAnglesPanel(bpy.types.Panel):
    bl_label = "Bone Angles Local"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'

    def draw(self, context):
        arm = get_armature()
 
        row = self.layout.row()
        row.label(text='Bone name')
        row.label(text='Degrees X')
        row.label(text='Degrees Y')
        row.label(text='Degrees Z')
        
        for bone in arm.pose.bones:
            row = self.layout.row()
            row.label(text=bone.name)
            if bone.parent:
                Original_rot = bone.matrix_basis.copy()

                a = bone.matrix.copy()
                b = bone.parent.matrix.copy()

                q1 = a.to_quaternion()

                q2 = b.to_quaternion()

                difQuat = q1.rotation_difference(q2)

                difEuler = difQuat.to_euler()
                mat = difEuler
            
            else:
                a = bone.matrix.copy()
                q1 = a.to_quaternion()
                mat = q1.to_euler()
                
            x = mat.x
            y = mat.y
            z = mat.z
                
            if x < 0.0001:
                x = 0.0
            if y < 0.0001:
                y = 0.0
            if z < 0.0001:
                z = 0.0
            row.label(text='X: {:.3}'.format(degrees(x)))
            row.label(text='Y: {:.3}'.format(degrees(y)))
            row.label(text='Z: {:.3}'.format(degrees(z)))
                

bpy.utils.register_class(LOCAL_PT_BoneAnglesPanel)


from rich.progress import track
import cv2
import imutils
import numpy as np
import os


def VideoTrim(session, vidList, ft, parameterDictionary, rotationState, numCamRange):
    """ 
    From the list of saved frames from the time-syncing function, create synced videos by pulling those specific 
    frames from the original raw videos into new videos, replacing any duplicate frames with a black 'buffer' slide
    """  
    camList = list(
        ft.columns[1 : len(vidList) + 1]
    )  # grab the camera identifiers from the data frame
    resWidth = parameterDictionary.get("resWidth")
    resHeight = parameterDictionary.get("resHeight")
    framerate = parameterDictionary.get("framerate")
    trueWidth = resWidth  # creating true resWidth and resHeight in case we need to rotate our images
    trueHeight = resHeight
    codec = parameterDictionary.get("codec")
    postTrimmingTotalNumFrames = []

    for vid,cam,camNum in zip(vidList,camList,numCamRange): #iterate in parallel through camera identifiers and matching videos
        this_vid_path = session.rawVidPath/vid
        print('synchronizing video at - ' + str(this_vid_path))
        cap = cv2.VideoCapture(str(this_vid_path)) #initialize OpenCV capture
        frameTable = ft[cam] #grab the frames needed for that camera
        success, image = cap.read() #start reading frames
        fourcc = cv2.VideoWriter_fourcc(*codec)
        saveName = session.sessionID + "_synced_" + cam + ".mp4"

        saveSyncedVidPath = str(
            session.syncedVidPath / saveName
        )  # create an output path for the function
        if rotationState[camNum] == 90 or rotationState[camNum] == 270:
            # if we are rotating to these angles, we need to temporarily swap our resWidth and resHeight
            tempHeight = resHeight
            resHeight = resWidth
            resWidth = tempHeight

        out = cv2.VideoWriter(
            saveSyncedVidPath, fourcc, framerate, (resWidth, resHeight)
        )  # change resolution as needed
        count = 0
        for frame in track(frameTable):  # start looking through the frames we need
            if frame == -1:  # this is a buffer frame
                blankFrame = np.zeros_like(image)  # create a blank frame
                if rotationState[camNum] is not None:
                    blankFrame = imutils.rotate_bound(
                        blankFrame, angle=rotationState[camNum]
                    )
                    blankFrame = cv2.resize(blankFrame, (resWidth, resHeight))
                out.write(blankFrame)  # write that frame to the video
                count += 1
            else:
                cap.set(
                    cv2.CAP_PROP_POS_FRAMES, frame
                )  # set the video to the frame that we need
                success, image = cap.read()
                if rotationState[camNum] is not None:
                    image = imutils.rotate_bound(image, angle=rotationState[camNum])
                    image = cv2.resize(image, (resWidth, resHeight))
                out.write(image)
                f = 2
            
        resWidth = trueWidth
        resHeight = trueHeight
        cap.release()
        out.release()
        frame_length_cap = cv2.VideoCapture(saveSyncedVidPath)
        thisCamNumFrame = int(frame_length_cap.get(cv2.CAP_PROP_FRAME_COUNT)) 
        postTrimmingTotalNumFrames.append(thisCamNumFrame)
        print('Saved '+ saveSyncedVidPath)
        print()

    assert postTrimmingTotalNumFrames.count(postTrimmingTotalNumFrames[0]) == len(postTrimmingTotalNumFrames), "Number of frames in each synced video is not the same"
    session.postTrimmingNumFrames = postTrimmingTotalNumFrames[0]
    session.numFrames = postTrimmingTotalNumFrames[0]


def createCalibrationVideos(session,calVideoFrameLength,parameterDictionary):

    vidList = os.listdir(session.syncedVidPath)
    framelist = list(range(calVideoFrameLength))
    codec = parameterDictionary.get("codec")
    for count, vid in enumerate(vidList, start=1):
        cam_name = "Cam{}".format(count)
        cap = cv2.VideoCapture(str(session.syncedVidPath / vid))
        fourcc = cv2.VideoWriter_fourcc(*codec)

        # grab resolution parameters from the videos
        resWidth = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        resHeight = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        framerate = int(cap.get(cv2.CAP_PROP_FPS))

        saveName = (
            session.sessionID + "_trimmed_" + cam_name + ".mp4"
        )  # create a name for the trimmed video
        saveCalVidPath = str(
            session.calVidPath / saveName
        )  # create an output path for the function

        success, image = cap.read()  # start reading frames

        out = cv2.VideoWriter(saveCalVidPath, fourcc, framerate, (resWidth, resHeight))
        print("Trimming " + cam_name)
        for frame in track(framelist):
            cap.set(
                cv2.CAP_PROP_POS_FRAMES, frame
            )  # set the video to the frame that we need
            success, image = cap.read()
            out.write(image)
    cap.release()
    out.release()



from rich.console import Console
from rich.layout import Layout

console = Console()
layout = Layout()

# Divide the "screen" in to three parts
layout.split(
    Layout(name="header", size=3),
    Layout(ratio=1, name="main"),
    Layout(size=10, name="footer"),
)
# Divide the "main" layout in to "side" and "body"
layout["main"].split(
    Layout(name="side"),
    Layout(name="body", ratio=2),
    direction="horizontal"
)
# Divide the "side" layout in to two
layout["side"].split(Layout(), Layout())

console.print(layout)


import logging
from typing import List

import cv2
from pydantic import BaseModel

CAM_CHECK_NUM = 5

logger = logging.getLogger(__name__)


class RawCamera(BaseModel):
    name: str
    index: int


class FindAvailableResponse(BaseModel):
    camera_found_count: int
    cams_to_use: List[RawCamera]


class DetectPossibleCameras:

    def find_available_cameras(self) -> FindAvailableResponse:
        capBackend = cv2.CAP_ANY

        cams_to_use_list = []
        for camNum in range(CAM_CHECK_NUM):
            cap = cv2.VideoCapture(camNum, capBackend)
            success, image = cap.read()
            if success:
                cams_to_use_list.append(RawCamera(
                    index=camNum,
                    name='test'
                ))
                cap.release()
                logger.info("I was successful")
                logger.info(cap.get(cv2.CAP_PROP_SETTINGS))

        return FindAvailableResponse(
            camera_found_count=len(cams_to_use_list),
            cams_to_use=cams_to_use_list
        )







import logging
import logging.handlers
import sys
from logging.config import dictConfig


DEFAULT_LOGGING = {"version": 1, "disable_existing_loggers": False}


def get_logging_handlers():
    """
    Initialize logging defaults for Project.
    :param logfile_path: logfile used to the logfile
    :type logfile_path: string
    This function does:
    - Assign INFO and DEBUG level to logger file handler and console handler
    """
    dictConfig(DEFAULT_LOGGING)

    default_formatter = logging.Formatter(
        "[%(asctime)s] [%(levelname)8s] [%(name)s] [%(funcName)s():%(lineno)s] [PID:%(process)d "
        "TID:%(thread)d] %(message)s",
        "%d/%m/%Y %H:%M:%S",
    )

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.DEBUG)
    console_handler.setFormatter(default_formatter)

    return [console_handler]


def configure_logging():
    handlers = get_logging_handlers()
    logging.getLogger("").handlers.extend(handlers)
    logging.root.setLevel(logging.DEBUG)


def disable_flask_logging(flask_app):
    flask_app.logger.handlers = []
    flask_app.logger.propagate = False


import os
import copy
import json
import pickle 

from aniposelib.cameras import CameraGroup
from aniposelib.utils import load_pose2d_fnames
from numba.core.types.misc import StringLiteral
import numpy as np
import cv2

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from freemocap import reconstruct3D, fmc_anipose
from rich.progress import track
from rich import print
from rich.console import Console
console = Console()


def CalibrateCaptureVolume(session,board, calVideoFrameLength = .5):
    """ 
    Check if a previous calibration yaml exists, and if not, create a set of shortened calibration videos and run Anipose functions
    to create a calibration yaml. Takes the 2D charuco board points and reconstructs them into 3D points that are saved out
    into the DataArrays folder

    note `debugAniposeIter` spoofs a failed anipose calibration to make it easier to test the thing where we iterate through with different frame windows to try to find one that works
    """  
    
    session.calVidPath.mkdir(exist_ok = True)
    session.dataArrayPath.mkdir(exist_ok = True)
    calibrationVideoPath = session.calVidPath

    if type(calVideoFrameLength)==int or type(calVideoFrameLength)==float:
        if calVideoFrameLength < 0: # if '-1' use the whole video
            cal_vid_frame_range = [0, session.numFrames]
            calibrationVideoPath = session.calVidPath
        else:
            if calVideoFrameLength>0 and calVideoFrameLength<1: #if between 0 and 1, use as a percentage of the total video length
                cal_vid_frame_range = [0, round(calVideoFrameLength * session.numFrames)]
            else: #otherwise, just use the input value as the number of frames to use        
                cal_vid_frame_range = [0, calVideoFrameLength]
    elif type(calVideoFrameLength)==list:
        if len(calVideoFrameLength) ==2:
            cal_vid_frame_range = calVideoFrameLength

    createCalibrationVideos(session, cal_vid_frame_range)



    vidnames = []
    cam_names = []
    for count, thisVidPath in enumerate(calibrationVideoPath.glob("*.mp4"), start=1):
        vidnames.append([str(thisVidPath)])
        cam_names.append(str(count))
        session.numCams = count

    cgroup = fmc_anipose.CameraGroup.from_names( cam_names, fisheye=True  )  # Looking through their code... it looks lke the 'fisheye=True' doesn't do much (see 2020-03-29 obsidian note)

    calibrationFile = "{}_calibration.toml".format(session.sessionID)

    session.cameraCalFilePath = session.sessionPath / calibrationFile

    error,charuco_data, charuco_frames = cgroup.calibrate_videos(vidnames, board)
    
    ## iteratively retry if anipose fails, butcouldn't get this to work because anipose launches some weird thread process that doesn't terminate properly? 
    # try:
    #     error,charuco_data, charuco_frames = cgroup.calibrate_videos(vidnames, board)
    #     if not spoof_anipose_fail_bool:
    #         anipose_success = True        
    #     else:
    #         anipose_success = False
    # except:
    #     anipose_success = False
    #     console.rule('Anipose Calibration Failed for frame range {} - {}'.format(calVideoFrameLength[0], calVideoFrameLength[1]), style="color({})".format(3))

    # anipose_iter = -1
    # if not anipose_success: #%% run anipose calibration, and if it fails, re-make calibration vids with different frame range and try again
        
    #     frame_window_size = round(session.numFrames*.5)
    #     cal_video_frame_range = [0, frame_window_size]
    #     while not anipose_success:
    #             anipose_iter += 1
    #             console.rule('Anipose Failed - Reprocessing - Iteration #{}'.format(anipose_iter), style="color({})".format(3))
                
    #             if cal_video_frame_range[1] > session.numFrames:
    #                 console.rule('We moved the window all the way through the videos and Anipose still isn\'t happy, starting again from the begining with a shorter frame window', style="color({})".format(3))
    #                 frame_window_size = round(frame_window_size/2)
    #                 cal_video_frame_range = [0, frame_window_size]
                                    
    #                 if round(cal_video_frame_range[-1] * session.numFrames) < 1 or anipose_iter > 20:
    #                     Exception('Sorry, we weren\'t able to get AniPose to run successfully :( - Make sure your Charuco board was visible to each camera (not necessarily at the same time) and that there wasn\'t, like, a bunch of glare on it')
    #                     return
            
    #             console.rule('Trying Anipose calibration again with frame range {} - {}'.format(cal_video_frame_range[0], cal_video_frame_range[1]), style="color({})".format(3))
    #             createCalibrationVideos(session, cal_video_frame_range)

    #             try:
    #                 error,charuco_data, charuco_frames = cgroup.calibrate_videos(vidnames, board)
                    
    #                 if not spoof_anipose_fail_bool:
    #                     anipose_success=True
    #                 else:
    #                     cal_video_frame_range[0] = cal_video_frame_range[0] + round(session.numFrames*.1)
    #                     cal_video_frame_range[1] = cal_video_frame_range[1] + round(session.numFrames*.1)
    #             except:
    #                 cal_video_frame_range[0] = cal_video_frame_range[0] + round(session.numFrames*.1)
    #                 cal_video_frame_range[1] = cal_video_frame_range[1] + round(session.numFrames*.1)


    cgroup.dump(session.cameraCalFilePath) 

    camera_calibration_info_dict = cgroup.get_dicts()
    camera_calibration_pickle_path = session.sessionPath / "{}_calibration.pickle".format(session.sessionID)
    
    with open(str(camera_calibration_pickle_path), 'wb') as pickle_file:
        pickle.dump(camera_calibration_info_dict, pickle_file)
        
    # camera_calibration_json_filename = "{}_calibration.json".format(session.sessionID)
    # camera_calibration_json_path = session.sessionPath / camera_calibration_json_filename
    
    # with open(camera_calibration_json_path, "w") as outfile:
    #     for camera_number, this_cam_calib_info in enumerate(camera_calibration_info_dict):
    #         camera_name = "camera_"+str(camera_number)
    #         this_cam_dict = {}
    #         this_cam_dict[camera_name] = this_cam_calib_info
    #         json.dumps(this_cam_dict, outfile)    


    session.cgroup = cgroup
    n_frames = cal_vid_frame_range[1]-cal_vid_frame_range[0]
    startframe = 0
    n_trackedPoints = 24

    charuco_nCams_nFrames_nImgPts_XY = np.empty([session.numCams, n_frames, n_trackedPoints,  2])
    charuco_nCams_nFrames_nImgPts_XY[:] = np.nan

    for cam in range(session.numCams):
        for charCount, thisCharFrame in enumerate(charuco_frames):
            try:
                charuco_nCams_nFrames_nImgPts_XY[cam, thisCharFrame, :,:] = np.squeeze(charuco_data[charCount][cam]["filled"])
            except:
                # print("failed frame:", frame)
                continue
    
    charuco2d_filename = session.dataArrayPath/'charuco_2d_points.npy'
    np.save(charuco2d_filename,charuco_nCams_nFrames_nImgPts_XY)

    session.charuco_nCams_nFrames_nImgPts_XY = charuco_nCams_nFrames_nImgPts_XY

    charuco_fr_mar_xyz, charuco_reprojErr= reconstruct3D.reconstruct3D(
        session, charuco_nCams_nFrames_nImgPts_XY
    )

    mean_charuco_fr_mar_xyz = np.nanmean(charuco_fr_mar_xyz, axis=0)

    # charry_reshaped = charucoarray.reshape(session.numCams, -1, 2)

    # char_flat = cgroup.triangulate(charry_reshaped, progress=True)
    # charReprojerr_flat = cgroup.reprojection_error(char_flat, charry_reshaped, mean=True)

    # char_fr_mar_xyz = char_flat.reshape(n_frames, n_trackedPoints, 3)
    # charReprojErr_fr_mar_err = charReprojerr_flat.reshape(n_frames, n_trackedPoints)

    if session.debug:
        fig = plt.figure()
        # mean charuco position
        ax1 = fig.add_subplot(111, projection="3d")
        ax1.cla()
        x = mean_charuco_fr_mar_xyz[:][:, 0]
        y = mean_charuco_fr_mar_xyz[:][:, 1]
        z = mean_charuco_fr_mar_xyz[:][:, 2]
        mx = np.nanmean(x)
        my = np.nanmean(y)
        mz = np.nanmean(z)

        ax1.scatter(x, y, z, marker="o")
        ax1.set_title("Mean Charuco Point Positions")

        ax1.set_xlabel("x")
        ax1.set_ylabel("y")
        ax1.set_zlabel("z")

        for camNum in range(len(cgroup.cameras)):
            np.append(mx, cgroup.cameras[camNum].tvec[0])
            np.append(my, cgroup.cameras[camNum].tvec[1])
            np.append(mz, cgroup.cameras[camNum].tvec[2])

            ax1.scatter(
                cgroup.cameras[camNum].tvec[0],
                cgroup.cameras[camNum].tvec[1],
                cgroup.cameras[camNum].tvec[2],
                marker="p",
            )

        axRange = board.square_length * 5

        ax1.set_xlim(mx - axRange, mx + axRange)
        ax1.set_ylim(my - axRange, my + axRange)
        ax1.set_zlim(mz - axRange, mz + axRange)

        # #charuco points over time
        # ax2 = fig.add_subplot(122)
        # ax2.cla()
        # numPts = charuco_fr_mar_xyz.shape[1]
        # for pp in range(numPts):
        #     ax2.plot(charuco_fr_mar_xyz[:,pp,:])
        # ax2.set_title('Charuco Point positions over time')
        plt.show()

    np.save(session.dataArrayPath/'charuco_3d_points.npy', charuco_fr_mar_xyz)
    np.save(session.dataArrayPath/'charuco_3d_reprojErr.npy', charuco_reprojErr)

    return cgroup, charuco_fr_mar_xyz


def createCalibrationVideos(session, calVideoFrameLength):
    """ 
    Based on the desired length of the calibration videos (for the anipose functions), create new videos trimmed 
    to that specific length
    """  
    vidList = os.listdir(session.syncedVidPath)
    if len(calVideoFrameLength)==1:
        framelist = list(range(calVideoFrameLength))
    elif len(calVideoFrameLength)==2:
        framelist = list(range(calVideoFrameLength[0], calVideoFrameLength[1]))
    else:
        Exception('calVideoFrameLength must be either 1 or 2 elements long')
    
    codec = "DIVX"
    for count, vid in enumerate(vidList, start=1):
        cam_name = "Cam{}".format(count)
        cap = cv2.VideoCapture(str(session.syncedVidPath / vid))
        fourcc = cv2.VideoWriter_fourcc(*codec)

        # grab resolution parameters from the videos
        resWidth = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        resHeight = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        framerate = int(cap.get(cv2.CAP_PROP_FPS))

        saveName = (
            session.sessionID + "_trimmed_" + cam_name + ".mp4"
        )  # create a name for the trimmed video
        saveCalVidPath = str(
            session.calVidPath / saveName
        )  # create an output path for the function

        success, image = cap.read()  # start reading frames

        out = cv2.VideoWriter(saveCalVidPath, fourcc, framerate, (resWidth, resHeight))
        print("Trimming " + cam_name + " to frames {}-{} for Anipose Calibration".format(framelist[0], framelist[-1]))
        for frame in track(framelist):
            cap.set(
                cv2.CAP_PROP_POS_FRAMES, frame
            )  # set the video to the frame that we need
            success, image = cap.read()
            out.write(image)
        cap.release()
        out.release()



from pathlib import Path

from pathos.helpers import mp as pathos_mp_helper
from rich.console import Console

from FMC_MultiCamera import FMC_MultiCamera
import freemocap as fmc



if __name__ == '__main__':
    pathos_mp_helper.freeze_support()
    console = Console() #create rich console to catch and print exceptions

    import socket
    this_computer_name = socket.gethostname()

    freemocap_data_path=None
    in_rotation_codes_list=None
        
    if this_computer_name=='jon-hallway-XPS-8930':
        freemocap_data_path = Path('/home/jon/Dropbox/FreeMoCapProject/FreeMocap_Data')
        in_rotation_codes_list = ['cv2.ROTATE_90_COUNTERCLOCKWISE', 'cv2.ROTATE_90_COUNTERCLOCKWISE', 'cv2.ROTATE_90_CLOCKWISE', 'cv2.ROTATE_90_CLOCKWISE', 'cv2.ROTATE_90_CLOCKWISE', ]
    elif this_computer_name == 'DESKTOP-DCG6K4F':
        freemocap_data_path = Path(r'C:\Users\jonma\Dropbox\FreeMoCapProject\FreeMocap_Data')


    try:
        multi_cam = FMC_MultiCamera(freemocap_data_folder=str(freemocap_data_path), rotation_codes_list=in_rotation_codes_list)
        multi_cam.start(standalone_mode=True)        
    except Exception:
        console.print_exception()
        
    fmc.RunMe(multi_cam._rec_name, stage=3)
    